from dataclasses import dataclass
from math import ceil, log

import numpy as np
import numpy.typing as npt

from .util import Word, integer_to_symbols, symbols_to_integer


@dataclass
class LempelZivWelchCode:
    r"""
    Lempel–Ziv–Welch (LZW) code. It is a lossless data compression algorithm which is variation of the [Lempel–Ziv 78](/ref/LempelZiv78Code) algorithm. For more details, see <cite>Say06, Sec. 5.4.2</cite>.

    Parameters:
        source_cardinality: The source cardinality $S$. Must be an integer greater than or equal to $2$.
        target_cardinality: The target cardinality $T$. Must be an integer greater than or equal to $2$. Default is $2$ (binary).

    Examples:
        >>> lzw = komm.LempelZivWelchCode(2)  # Binary source, binary target
        >>> lzw = komm.LempelZivWelchCode(3, 4)  # Ternary source, quaternary target
    """

    source_cardinality: int
    target_cardinality: int = 2

    def __post_init__(self) -> None:
        if not self.source_cardinality >= 2:
            raise ValueError("'source_cardinality': must be at least 2")
        if not self.target_cardinality >= 2:
            raise ValueError("'target_cardinality': must be at least 2")

    def encode(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Encodes a sequence of source symbols using the LZW encoding algorithm.

        Parameters:
            input: The sequence of symbols to be encoded. Must be a 1D-array with elements in $[0:S)$ (where $S$ is the source cardinality of the code).

        Returns:
            output: The sequence of encoded symbols. It is a 1D-array with elements in $[0:T)$ (where $T$ is the target cardinality of the code).

        Examples:
            >>> lzw = komm.LempelZivWelchCode(2)
            >>> lzw.encode(np.zeros(15, dtype=int))
            array([0, 1, 0, 1, 1, 1, 0, 0, 1, 0, 1])

            >>> lzw = komm.LempelZivWelchCode(2, 8)
            >>> lzw.encode(np.zeros(15, dtype=int))
            array([0, 2, 3, 4, 5])
        """
        T = self.target_cardinality
        S = self.source_cardinality
        input = np.asarray(input)
        dictionary: dict[Word, int] = {(s,): s for s in range(S)}
        output: list[int] = []

        word: Word = ()
        for symbol in input:
            if word + (symbol,) in dictionary:
                word += (symbol,)
                continue
            k = ceil(log(len(dictionary), T))
            pointer = dictionary[word]
            output.extend(integer_to_symbols(pointer, base=T, width=k))
            dictionary[word + (symbol,)] = len(dictionary)
            word = (symbol,)

        if word:
            k = ceil(log(len(dictionary), T))
            pointer = dictionary[word]
            output.extend(integer_to_symbols(pointer, base=T, width=k))

        return np.array(output, dtype=int)

    def decode(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Decodes a sequence of encoded symbols using the LZW decoding algorithm.

        Parameters:
            input: The sequence of symbols to be decoded. Must be a 1D-array with elements in $[0:T)$ (where $T$ is the target cardinality of the code). Also, the sequence must be a valid output of the `encode` method.

        Returns:
            output: The sequence of decoded symbols. It is a 1D-array with elements in $[0:S)$ (where $S$ is the source cardinality of the code).

        Examples:
            >>> lzw = komm.LempelZivWelchCode(2)
            >>> lzw.decode([0, 1, 0, 1, 1, 1, 0, 0, 1, 0, 1])
            array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])

            >>> lzw = komm.LempelZivWelchCode(2, 8)
            >>> lzw.decode([0, 2, 3, 4, 5])
            array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])
        """
        T = self.target_cardinality
        S = self.source_cardinality
        input = np.asarray(input)

        if input.size == 0:
            return np.array([], dtype=int)

        dictionary: dict[int, Word] = {s: (s,) for s in range(S)}
        output: list[int] = []

        k = ceil(log(S, T))
        pointer = symbols_to_integer(input[:k], base=T)
        old = dictionary[pointer]
        output.extend(old)

        i = k
        while True:
            k = ceil(log(len(dictionary) + 1, T))
            if i + k > input.size:
                break
            pointer = symbols_to_integer(input[i : i + k], base=T)
            word = dictionary.get(pointer, old + (old[0],))
            output.extend(word)
            dictionary[len(dictionary)] = old + (word[0],)
            old = word
            i += k

        return np.array(output, dtype=int)



from typing import Protocol, TypeVar, runtime_checkable

from typing_extensions import Self

from . import ring

T_co = TypeVar("T_co", bound="FieldElement", covariant=True)


@runtime_checkable
class FieldElement(ring.RingElement, Protocol):
    def inverse(self: Self) -> Self: ...
    def __truediv__(self: Self, other: Self) -> Self: ...


def power(x: T_co, n: int) -> T_co:
    """Compute $x^n$ using exponentiation by squaring. See the corresponding function in :mod:`komm._algebra.ring`.

    Parameters:
        x: The base (a field element)
        n: The exponent

    Returns:
        power: The result of `x` raised to the power of `n` in the field
    """
    if n < 0:
        return ring.power(x.inverse(), -n)
    else:
        return ring.power(x, n)



from dataclasses import dataclass
from functools import cached_property
from typing import Literal

import numpy as np
import numpy.typing as npt

from . import base


@dataclass(eq=False)
class TransmitFilter:
    r"""
    Transmit filter (pulse shaping). Given a sequence of $N$ real or complex symbols $x[n]$, this filter outputs samples of the signal
    $$
        x(t) = \sum_{n=0}^{N-1} x[n] h(t - n),
    $$
    where $h(t)$ is the waveform of a given [pulse](/ref/Pulse), and the samples of the output signal are taken at an integer rate of $\beta$ samples per symbol. Note that the symbol interval is normalized to $1$.

    The time span of $x(t)$ is given by $[ n_0, n_1 + N - 1 )$, where $[ n_0, n_1 )$ is the integer-bounded time span of $h(t)$. In turn, $n_0$ and $n_1$ depend on the support of $h(t)$:

    - If $h(t)$ has finite support $[ t_0, t_1 ]$, then $n_0 = \lfloor t_0 \rfloor$ and $n_1 = \lceil t_1 \rceil$.

    - If $h(t)$ has infinite support, then $n_0 = -L/2$ and $n_1 = L/2$, where $L$ is a given even positive integer, called the _truncation window length_.

    Attributes:
        pulse: The pulse whose waveform is $h(t)$.
        samples_per_symbol: The number $\beta$ of samples (of the output) per symbol (of the input). Must be a positive integer.
        truncation: The truncation window length $L$. Only applies to infinite-duration pulses. Must be an even positive integer. The default value is `32`.
    """

    pulse: base.Pulse
    samples_per_symbol: int
    truncation: int | None = None

    def _pulse_support_kind(self) -> Literal["finite", "infinite", "semi-infinite"]:
        support = self.pulse.support
        if support == (-np.inf, np.inf):
            return "infinite"
        elif support[0] != -np.inf and support[1] != np.inf:
            return "finite"
        else:
            return "semi-infinite"

    def __post_init__(self) -> None:
        if self._pulse_support_kind() == "finite":
            if self.truncation is not None:
                raise ValueError("'truncation' only applies to infinite-support pulses")
        elif self._pulse_support_kind() == "infinite":
            if self.truncation is None:
                self.truncation = 32
            elif self.truncation <= 0 or self.truncation % 2 != 0:
                raise ValueError("'truncation' must be an even positive integer")
        else:  # self._pulse_support_kind() == "semi-infinite"
            raise ValueError("pulses with semi-infinite support are not supported")

    @cached_property
    def pulse_time_span(self) -> tuple[int, int]:
        r"""
        The integer-bounded time span $[ n_0, n_1 )$ of the pulse waveform $h(t)$.

        Examples:
            >>> pulse = komm.RectangularPulse(0.25)
            >>> pulse.support
            (0.0, 0.25)
            >>> tx_filter = komm.TransmitFilter(pulse=pulse, samples_per_symbol=3)
            >>> tx_filter.pulse_time_span
            (0, 1)

            >>> pulse = komm.SincPulse()
            >>> pulse.support
            (-inf, inf)
            >>> tx_filter = komm.TransmitFilter(pulse=pulse, samples_per_symbol=4, truncation=4)
            >>> tx_filter.pulse_time_span
            (-2, 2)
        """
        if self._pulse_support_kind() == "infinite":
            assert self.truncation is not None  # Guaranteed
            t_min_h = -self.truncation // 2
            t_max_h = self.truncation // 2
        else:  # self._pulse_support_kind() == "finite"
            support = self.pulse.support
            t_min_h = int(np.floor(support[0]))
            t_max_h = int(np.ceil(support[1]))
        return t_min_h, t_max_h

    def _time(self, num_symbols: int) -> npt.NDArray[np.floating]:
        # The time axis of the output signal considering 'num_symbols' input symbols.
        t_min_h, t_max_h = self.pulse_time_span
        t_min_x, t_max_x = t_min_h, t_max_h + num_symbols - 1
        beta = self.samples_per_symbol
        t = np.arange(t_min_x * beta, t_max_x * beta) / beta
        return t

    @cached_property
    def taps(self) -> npt.NDArray[np.floating]:
        r"""
        The FIR filter taps of the transmit filter.

        Examples:
            >>> pulse = komm.RectangularPulse(width=0.25)
            >>> tx_filter = komm.TransmitFilter(pulse=pulse, samples_per_symbol=3)
            >>> tx_filter.taps
            array([1., 0., 0.])

            >>> pulse = komm.SincPulse()
            >>> tx_filter = komm.TransmitFilter(pulse=pulse, samples_per_symbol=4, truncation=4)
            >>> tx_filter.taps.reshape((-1, 4)).round(6)
            array([[-0.      , -0.128617, -0.212207, -0.180063],
                   [ 0.      ,  0.300105,  0.63662 ,  0.900316],
                   [ 1.      ,  0.900316,  0.63662 ,  0.300105],
                   [ 0.      , -0.180063, -0.212207, -0.128617]])
        """
        return self.pulse.waveform(self._time(1))

    def time(self, input: npt.ArrayLike) -> npt.NDArray[np.floating]:
        r"""
        Convenience function to generate the time axis of the output signal given the input symbols.

        Parameters:
            input: The input symbols $x[n]$, of length $N$.

        Returns:
            t: The time axis of the output signal, of length $(N + n_1 - n_0 - 1) \beta$.

        Examples:
            >>> pulse = komm.RectangularPulse()
            >>> tx_filter = komm.TransmitFilter(pulse=pulse, samples_per_symbol=3)
            >>> tx_filter.time([1.0, -1.0, 1.0, 1.0]).round(2)
            array([0.  , 0.33, 0.67, 1.  , 1.33, 1.67, 2.  , 2.33, 2.67, 3.  , 3.33, 3.67])

            >>> pulse = komm.SincPulse()
            >>> tx_filter = komm.TransmitFilter(pulse=pulse, samples_per_symbol=4, truncation=4)
            >>> tx_filter.time([1.0, -1.0, 1.0, 1.0]).reshape((-1, 4))
            array([[-2.  , -1.75, -1.5 , -1.25],
                   [-1.  , -0.75, -0.5 , -0.25],
                   [ 0.  ,  0.25,  0.5 ,  0.75],
                   [ 1.  ,  1.25,  1.5 ,  1.75],
                   [ 2.  ,  2.25,  2.5 ,  2.75],
                   [ 3.  ,  3.25,  3.5 ,  3.75],
                   [ 4.  ,  4.25,  4.5 ,  4.75]])
        """
        input = np.asarray(input)
        return self._time(input.size)

    def __call__(
        self, input: npt.ArrayLike
    ) -> npt.NDArray[np.floating | np.complexfloating]:
        r"""
        Process the input symbols through the transmit filter.

        Parameters:
            input: The input symbols $x[n]$, of length $N$.

        Returns:
            output: The samples of the output signal $x(t)$, of length $(N + n_1 - n_0 - 1) \beta$.

        Examples:
            >>> pulse = komm.RectangularPulse(width=1.0)
            >>> tx_filter = komm.TransmitFilter(pulse=pulse, samples_per_symbol=3)
            >>> tx_filter([1.0, -1.0, 1.0, 1.0])
            array([ 1.,  1.,  1., -1., -1., -1.,  1.,  1.,  1.,  1.,  1.,  1.])

            >>> pulse = komm.RectangularPulse(width=0.25)
            >>> tx_filter = komm.TransmitFilter(pulse=pulse, samples_per_symbol=3)
            >>> tx_filter([1.0, -1.0, 1.0, 1.0])
            array([ 1.,  0.,  0., -1.,  0.,  0.,  1.,  0.,  0.,  1.,  0.,  0.])

            >>> pulse = komm.SincPulse()
            >>> tx_filter = komm.TransmitFilter(pulse=pulse, samples_per_symbol=4, truncation=4)
            >>> tx_filter([1.0, -1.0, 1.0, 1.0]).reshape((-1, 4)).round(6)
            array([[-0.      , -0.128617, -0.212207, -0.180063],
                   [ 0.      ,  0.428722,  0.848826,  1.08038 ],
                   [ 1.      ,  0.471594, -0.212207, -0.780274],
                   [-1.      , -0.908891, -0.424413,  0.291531],
                   [ 1.      ,  1.380485,  1.485446,  1.329038],
                   [ 1.      ,  0.720253,  0.424413,  0.171489],
                   [ 0.      , -0.180063, -0.212207, -0.128617]])

            >>> pulse = komm.RectangularPulse()
            >>> tx_filter = komm.TransmitFilter(pulse=pulse, samples_per_symbol=4, truncation=4)
            Traceback (most recent call last):
            ...
            ValueError: 'truncation' only applies to infinite-support pulses
        """
        input = np.asarray(input)
        beta = self.samples_per_symbol
        input_interp = np.zeros((input.size - 1) * beta + 1, dtype=input.dtype)
        input_interp[::beta] = input
        output = np.convolve(self.taps, input_interp)
        return output



from dataclasses import dataclass, field
from functools import cached_property
from typing import Literal

import numpy as np
import numpy.typing as npt


@dataclass
class AWGNChannel:
    r"""
    Additive white Gaussian noise (AWGN) channel. It is defined by
    $$
        Y_n = X_n + Z_n,
    $$
    where $X_n$ is the channel *input signal*, $Y_n$ is the channel *output signal*, and $Z_n$ is the *noise*, which is iid according to a Gaussian distribution with zero mean. The channel *signal-to-noise ratio* is calculated by
    $$
        \snr = \frac{P}{N},
    $$
    where $P = \mathrm{E}[X^2_n]$ is the average power of the input signal, and $N = \mathrm{E}[Z^2_n]$ is the average power (and variance) of the noise. For more details, see <cite>CT06, Ch. 9</cite>.

    Attributes:
        signal_power: The input signal power $P$. If equal to the string `'measured'`, then every time the channel is invoked the input signal power will be computed from the input itself (i.e., its squared Euclidean norm).

        snr: The channel signal-to-noise ratio $\snr$ (linear, not decibel). The default value is `np.inf`, which corresponds to a noiseless channel.
    """

    signal_power: float | Literal["measured"]
    snr: float = np.inf
    rng: np.random.Generator = field(default=np.random.default_rng(), repr=False)

    @cached_property
    def noise_power(self) -> float:
        r"""
        The noise power $N$.
        """
        if self.signal_power == "measured":
            raise ValueError(
                "noise power cannot be calculated when 'signal_power' is 'measured'"
            )
        return self.signal_power / self.snr

    def capacity(self) -> float:
        r"""
        Returns the channel capacity $C$. It is given by $C = \frac{1}{2}\log_2(1 + \snr)$, in bits per dimension.

        Examples:
            >>> awgn = komm.AWGNChannel(signal_power=1.0, snr=63.0)
            >>> awgn.capacity()
            np.float64(3.0)
        """
        return 0.5 * np.log1p(self.snr) / np.log(2.0)

    def __call__(self, input: npt.ArrayLike) -> npt.NDArray[np.floating]:
        r"""
        Transmits the input signal through the channel and returns the output signal.

        Parameters:
            input: The input signal $X_n$.

        Returns:
            output: The output signal $Y_n$.

        Examples:
            >>> rng = np.random.default_rng(seed=42)
            >>> awgn = komm.AWGNChannel(signal_power=5.0, snr=200.0, rng=rng)
            >>> x = [1.0, 3.0, -3.0, -1.0, -1.0, 1.0, 3.0, 1.0, -1.0, 3.0]
            >>> awgn(x).round(2)
            array([ 1.05,  2.84, -2.88, -0.85, -1.31,  0.79,  3.02,  0.95, -1.  ,  2.87])
        """
        input = np.array(input)

        if self.signal_power == "measured":
            signal_power = np.linalg.norm(input) ** 2 / input.shape[-1]
        else:
            signal_power = self.signal_power

        noise_power = signal_power / self.snr

        if input.dtype == complex:
            noise = np.sqrt(noise_power / 2) * (
                self.rng.standard_normal(size=input.shape)
                + 1j * self.rng.standard_normal(size=input.shape)
            )
        else:
            noise = np.sqrt(noise_power) * self.rng.standard_normal(size=input.shape)

        return input + noise



from functools import cached_property
from typing import Literal

import numpy as np
import numpy.typing as npt

from .._util.decorators import vectorize
from .._util.special_functions import logcosh
from . import base
from .constellations import constellation_pam
from .labelings import get_labeling


class PAModulation(base.Modulation[np.floating]):
    r"""
    Pulse-amplitude modulation (PAM). It is a real [modulation scheme](/ref/Modulation) in which the constellation symbols are *uniformly arranged* in the real line and have zero mean. More precisely, the the $i$-th constellation symbol is given by
    $$
        x_i = A \left( 2i - M + 1 \right), \quad i \in [0 : M),
    $$
    where $M$ is the *order* (a power of $2$), and $A$ is the *base amplitude* of the modulation.

    Parameters:
        order: The order $M$ of the modulation. It must be a power of $2$.

        base_amplitude: The base amplitude $A$ of the constellation. The default value is `1.0`.

        labeling: The binary labeling of the modulation. Can be specified either as a 2D-array of integers (see [base class](/ref/Modulation) for details), or as a string. In the latter case, the string must be either `'natural'` or `'reflected'`. The default value is `'reflected'`, corresponding to the Gray labeling.

    Examples:
        1. The $4$-PAM modulation with base amplitude $A = 1$ and natural labeling is depicted below.
            <figure markdown>
            ![4-PAM modulation with Gray labeling.](/figures/pam_4_natural.svg)
            </figure>

                >>> pam = komm.PAModulation(4, labeling="natural")
                >>> pam.constellation
                array([-3., -1.,  1.,  3.])
                >>> pam.labeling
                array([[0, 0],
                       [0, 1],
                       [1, 0],
                       [1, 1]])

        1. The $8$-PAM modulation with base amplitude $A = 0.5$ and Gray labeling is depicted below.
            <figure markdown>
            ![8-PAM modulation with Gray labeling.](/figures/pam_8_gray.svg)
            </figure>

                >>> pam = komm.PAModulation(8, base_amplitude=0.5)
                >>> pam.constellation
                array([-3.5, -2.5, -1.5, -0.5,  0.5,  1.5,  2.5,  3.5])
                >>> pam.labeling
                array([[0, 0, 0],
                       [0, 0, 1],
                       [0, 1, 1],
                       [0, 1, 0],
                       [1, 1, 0],
                       [1, 1, 1],
                       [1, 0, 1],
                       [1, 0, 0]])
    """

    def __init__(
        self,
        order: int,
        base_amplitude: float = 1.0,
        labeling: Literal["natural", "reflected"] | npt.ArrayLike = "reflected",
    ) -> None:
        self._order = order
        self._base_amplitude = base_amplitude
        self._labeling = labeling
        super()._validate_parameters()

    def __repr__(self) -> str:
        if isinstance(self._labeling, str):
            labeling_repr = repr(self._labeling)
        else:
            labeling_repr = self.labeling.tolist()
        args = ", ".join([
            f"order={self._order}",
            f"base_amplitude={self._base_amplitude}",
            f"labeling={labeling_repr}",
        ])
        return f"{self.__class__.__name__}({args})"

    @cached_property
    def constellation(self) -> npt.NDArray[np.floating]:
        r"""
        Examples:
            >>> pam = komm.PAModulation(4)
            >>> pam.constellation
            array([-3., -1.,  1.,  3.])
        """
        return constellation_pam(self._order, self._base_amplitude)

    @cached_property
    def labeling(self) -> npt.NDArray[np.integer]:
        r"""
        Examples:
            >>> pam = komm.PAModulation(4)
            >>> pam.labeling
            array([[0, 0],
                   [0, 1],
                   [1, 1],
                   [1, 0]])
        """
        return get_labeling(self._labeling, ("natural", "reflected"), self._order)

    @cached_property
    def inverse_labeling(self) -> dict[tuple[int, ...], int]:
        r"""
        Examples:
            >>> pam = komm.PAModulation(4)
            >>> pam.inverse_labeling
            {(0, 0): 0, (0, 1): 1, (1, 1): 2, (1, 0): 3}
        """
        return super().inverse_labeling

    @cached_property
    def order(self) -> int:
        r"""
        Examples:
            >>> pam = komm.PAModulation(4)
            >>> pam.order
            4
        """
        return self._order

    @cached_property
    def bits_per_symbol(self) -> int:
        r"""
        Examples:
            >>> pam = komm.PAModulation(4)
            >>> pam.bits_per_symbol
            2
        """
        return super().bits_per_symbol

    @cached_property
    def energy_per_symbol(self) -> float:
        r"""
        For the PAM, it is given by
        $$
            E_\mathrm{s} = \frac{A^2}{3}(M^2 - 1).
        $$

        Examples:
            >>> pam = komm.PAModulation(4)
            >>> pam.energy_per_symbol
            5.0
        """
        return (self._base_amplitude**2) * (self._order**2 - 1) / 3

    @cached_property
    def energy_per_bit(self) -> float:
        r"""
        Examples:
            >>> pam = komm.PAModulation(4)
            >>> pam.energy_per_bit
            2.5
        """
        return super().energy_per_bit

    @cached_property
    def symbol_mean(self) -> float:
        r"""
        For the PAM, it is given by
        $$
            \mu_\mathrm{s} = 0.
        $$

        Examples:
            >>> pam = komm.PAModulation(4)
            >>> pam.symbol_mean
            0.0
        """
        return 0.0

    @cached_property
    def minimum_distance(self) -> float:
        r"""
        For the PAM, it is given by
        $$
            d_\mathrm{min} = 2A.
        $$

        Examples:
            >>> pam = komm.PAModulation(4)
            >>> pam.minimum_distance
            2.0
        """
        return 2.0 * self._base_amplitude

    def modulate(self, input: npt.ArrayLike) -> npt.NDArray[np.floating]:
        r"""
        Examples:
            >>> pam = komm.PAModulation(4)
            >>> pam.modulate([0, 0, 1, 1, 0, 0, 0, 1])
            array([-3.,  1., -3., -1.])
        """
        return super().modulate(input)

    def demodulate_hard(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        M, A = self._order, self._base_amplitude
        input = np.asarray(input) / A
        indices = np.clip(np.around((input + M - 1) / 2), 0, M - 1).astype(int)
        hard_bits = np.reshape(self.labeling[indices], shape=-1)
        return hard_bits

    def demodulate_soft(
        self, input: npt.ArrayLike, snr: float = 1.0
    ) -> npt.NDArray[np.floating]:
        if self._order == 2:
            y = np.asarray(input) / self._base_amplitude
            return self._demodulate_pam2_soft(y, snr / 1.0)
        elif self._order == 4 and self._labeling == "reflected":
            y = np.asarray(input) / self._base_amplitude
            return vectorize(self._demodulate_pam4_soft_reflected)(y, snr / 5.0)  # type: ignore
        # Fall back to general implementation
        return super().demodulate_soft(input, snr)

    def _demodulate_pam2_soft(self, y: npt.NDArray[np.floating], gamma: float):
        return -4 * gamma * y  # [SA15, eq. (3.65)]

    def _demodulate_pam4_soft_reflected(
        self, y: npt.NDArray[np.floating], gamma: float
    ):
        soft_bits = np.empty(2 * y.size, dtype=float)
        soft_bits[0::2] = (  # For bit_0: [SA15, eq. (5.6)]
            -8 * gamma * y + logcosh(2 * gamma * (y + 2)) - logcosh(2 * gamma * (y - 2))
        )
        soft_bits[1::2] = (  # For bit_1: [SA15, eq. (5.15)]
            -8 * gamma + logcosh(6 * gamma * y) - logcosh(2 * gamma * y)
        )
        return soft_bits



from functools import cached_property

import numpy as np
import numpy.typing as npt

from .._util.correlation import autocorrelation, cyclic_autocorrelation


class ComplexSequence:
    r"""
    General complex sequence. It is denoted by $x[n]$, with elements in $\\mathbb{C}$. Its length (or period) is denoted by $L$.

    Parameters:
        sequence: The complex sequence. Must be a 1D-array of length $L$ with elements in $\\mathbb{C}$.

    Examples:
        >>> seq = ComplexSequence([1, 1j, -1, -1j])
        >>> seq.sequence
        array([ 1.+0.j,  0.+1.j, -1.+0.j, -0.-1.j])
    """

    def __init__(self, sequence: npt.ArrayLike) -> None:
        self.sequence = np.asarray(sequence, dtype=complex)

    def __repr__(self) -> str:
        args = f"sequence={self.sequence.tolist()}"
        return f"{self.__class__.__name__}({args})"

    @cached_property
    def length(self):
        r"""
        The length (or period) $L$ of the complex sequence.
        """
        return self.sequence.size

    def autocorrelation(
        self, shifts: npt.ArrayLike | None = None, normalized: bool = False
    ) -> npt.NDArray[np.complexfloating]:
        r"""
        Returns the autocorrelation $R[\ell]$ of the complex sequence. See [`komm.autocorrelation`](/ref/autocorrelation) for more details.

        Parameters:
            shifts: See the corresponding parameter in [`komm.autocorrelation`](/ref/autocorrelation).
            normalized: See the corresponding parameter in [`komm.autocorrelation`](/ref/autocorrelation).

        Returns:
            The autocorrelation $R[\ell]$ of the complex sequence.

        Examples:
            >>> seq = ComplexSequence([1, 1j, -1, -1j])
            >>> seq.autocorrelation(shifts=[-2, -1, 0, 1, 2])
            array([-2.+0.j,  0.-3.j,  4.+0.j,  0.+3.j, -2.+0.j])
        """
        return autocorrelation(self.sequence, shifts, normalized)

    def cyclic_autocorrelation(
        self, shifts: npt.ArrayLike | None = None, normalized: bool = False
    ) -> npt.NDArray[np.complexfloating]:
        r"""
        Returns the cyclic autocorrelation $\tilde{R}[\ell]$ of the complex sequence. See [`komm.cyclic_autocorrelation`](/ref/cyclic_autocorrelation) for more details.

        Parameters:
            shifts: See the corresponding parameter in [`komm.cyclic_autocorrelation`](/ref/cyclic_autocorrelation).
            normalized: See the corresponding parameter in [`komm.cyclic_autocorrelation`](/ref/cyclic_autocorrelation).

        Returns:
            The cyclic autocorrelation $\tilde{R}[\ell]$ of the complex sequence.

        Examples:
            >>> seq = ComplexSequence([1, 1j, -1, -1j])
            >>> seq.cyclic_autocorrelation(shifts=[-2, -1, 0, 1, 2])
            array([-4.+0.j,  0.-4.j,  4.+0.j,  0.+4.j, -4.+0.j])
        """
        return cyclic_autocorrelation(self.sequence, shifts, normalized)



from importlib import metadata as _metadata

from ._algebra import *
from ._channels import *
from ._error_control_block import *
from ._error_control_checksum import *
from ._error_control_convolutional import *
from ._error_control_decoders import *
from ._finite_state_machine import *
from ._integer_coding import *
from ._lossless_coding import *
from ._modulation import *
from ._pulses import *
from ._quantization import *
from ._sequences_binary import *
from ._sequences_complex import *
from ._sources import *
from ._util import *

__version__ = _metadata.version("komm")



from dataclasses import dataclass
from functools import cached_property

import numpy as np
import numpy.typing as npt

from . import base


@dataclass
class SincPulse(base.Pulse):
    r"""
    Sinc pulse. It is a [pulse](/ref/Pulse) with waveform given by
    $$
        h(t) = \operatorname{sinc}(t) = \frac{\sin(\pi t)}{\pi t},
    $$
    and spectrum given by
    $$
        \hat{h}(f) = \begin{cases}
            1, & |f| < \frac{1}{2}, \\\\
            0, & \text{otherwise}.
        \end{cases}
    $$

    The sinc pulse is depicted below.

    <figure markdown>
      ![Sinc pulse.](/figures/pulse_sinc.svg)
    </figure>

    For more details, see <cite>PS08, Sec. 9.2-1</cite>.

    **Attributes:**

    <span style="font-size: 90%; font-style: italic; color: gray; margin-left: 1em;">(No attributes)</span>

    Examples:
        >>> pulse = komm.SincPulse()
        >>> pulse.waveform([-0.75, -0.50, -0.25,  0.00,  0.25,  0.50,  0.75]).round(4)
        array([0.3001, 0.6366, 0.9003, 1.    , 0.9003, 0.6366, 0.3001])
        >>> pulse.spectrum([-0.75, -0.50, -0.25,  0.00,  0.25,  0.50,  0.75])
        array([0., 0., 1., 1., 1., 0., 0.])
    """

    def waveform(self, t: npt.ArrayLike) -> npt.NDArray[np.floating]:
        t = np.asarray(t)
        return np.sinc(t)

    def spectrum(self, f: npt.ArrayLike) -> npt.NDArray[np.floating]:
        f = np.asarray(f)
        return 1.0 * (abs(f) < 0.5)

    @cached_property
    def support(self) -> tuple[float, float]:
        return (-np.inf, np.inf)



from dataclasses import dataclass, field
from functools import cached_property

import numpy as np
import numpy.typing as npt

from .._util.information_theory import (
    PMF,
    LogBase,
    assert_is_probability,
    binary_entropy,
)
from . import base


@dataclass
class BinarySymmetricChannel(base.DiscreteMemorylessChannel):
    r"""
    Binary symmetric channel (BSC). It is a [discrete memoryless channel](/ref/DiscreteMemorylessChannel) with input and output alphabets $\mathcal{X} = \mathcal{Y} = \\{ 0, 1 \\}$. The channel is characterized by a parameter $p$, called the *crossover probability*. With probability $1 - p$, the output symbol is identical to the input symbol, and with probability $p$, the output symbol is flipped. Equivalently, the channel can be modeled as
    $$
        Y_n = X_n + Z_n,
    $$
    where $Z_n$ are iid Bernoulli random variables with $\Pr[Z_n = 1] = p$. For more details, see <cite>CT06, Sec. 7.1.4</cite>.

    Attributes:
        crossover_probability: The channel crossover probability $p$. Must satisfy $0 \leq p \leq 1$. The default value is `0.0`, which corresponds to a noiseless channel.
    """

    crossover_probability: float = 0.0
    rng: np.random.Generator = field(default=np.random.default_rng(), repr=False)

    def __post_init__(self) -> None:
        assert_is_probability(self.crossover_probability)

    @cached_property
    def input_cardinality(self) -> int:
        r"""
        For the BSC, it is given by $|\mathcal{X}| = 2$.
        """
        return 2

    @cached_property
    def output_cardinality(self) -> int:
        r"""
        For the BSC, it is given by $|\mathcal{Y}| = 2$.
        """
        return 2

    @cached_property
    def transition_matrix(self) -> npt.NDArray[np.floating]:
        r"""
        For the BSC, it is given by
        $$
            p_{Y \mid X} = \begin{bmatrix} 1-p & p \\\\ p & 1-p \end{bmatrix}.
        $$

        Examples:
            >>> bsc = komm.BinarySymmetricChannel(0.2)
            >>> bsc.transition_matrix
            array([[0.8, 0.2],
                   [0.2, 0.8]])
        """
        p = self.crossover_probability
        return np.array([[1 - p, p], [p, 1 - p]])

    def mutual_information(
        self, input_pmf: npt.ArrayLike, base: LogBase = 2.0
    ) -> float:
        r"""
        For the BSC, it is given by
        $$
            \mathrm{I}(X ; Y) = \Hb(p + \pi - 2 p \pi) - \Hb(p),
        $$
        in bits, where $\pi = \Pr[X = 1]$, and $\Hb$ is the [binary entropy function](/ref/binary_entropy).

        Examples:
            >>> bsc = komm.BinarySymmetricChannel(0.2)
            >>> bsc.mutual_information([0.45, 0.55])
            np.float64(0.2754734936803773)
        """
        input_pmf = PMF(input_pmf)
        p = self.crossover_probability
        pi = input_pmf[1]
        return (binary_entropy(p + pi - 2 * p * pi) - binary_entropy(p)) / np.log2(base)

    def capacity(self, base: LogBase = 2.0) -> float:
        r"""
        For the BSC, it is given by
        $$
            C = 1 - \Hb(p),
        $$
        in bits, where $\Hb$ is the [binary entropy function](/ref/binary_entropy).

        Examples:
            >>> bsc = komm.BinarySymmetricChannel(0.2)
            >>> bsc.capacity()
            np.float64(0.2780719051126377)
        """
        p = self.crossover_probability
        return (1.0 - binary_entropy(p)) / np.log2(base)

    def __call__(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Examples:
            >>> rng = np.random.default_rng(seed=42)
            >>> bsc = komm.BinarySymmetricChannel(0.2, rng=rng)
            >>> bsc([1, 1, 1, 0, 0, 0, 1, 0, 1, 0])
            array([1, 1, 1, 0, 1, 0, 1, 0, 0, 0])
        """
        p = self.crossover_probability
        input = np.array(input)
        error_pattern = self.rng.random(input.shape) < p
        return (input + error_pattern) % 2



from .GaussianPulse import GaussianPulse
from .ManchesterPulse import ManchesterPulse
from .Pulse import Pulse
from .RaisedCosinePulse import RaisedCosinePulse
from .ReceiveFilter import ReceiveFilter
from .RectangularPulse import RectangularPulse
from .RootRaisedCosinePulse import RootRaisedCosinePulse
from .SincPulse import SincPulse
from .TransmitFilter import TransmitFilter

__all__ = [
    "GaussianPulse",
    "ManchesterPulse",
    "Pulse",
    "RaisedCosinePulse",
    "ReceiveFilter",
    "RectangularPulse",
    "RootRaisedCosinePulse",
    "SincPulse",
    "TransmitFilter",
]



from dataclasses import dataclass

import numpy as np
import numpy.typing as npt

from .._error_control_block.base import BlockCode
from .._util import bits_to_int
from .._util.decorators import blockwise
from . import base


@dataclass
class SyndromeTableDecoder(base.BlockDecoder[BlockCode]):
    r"""
    Syndrome table decoder for general [block codes](/ref/BlockCode). This decoder implements syndrome-based hard-decision decoding using a precomputed table of coset leaders.

    Parameters:
        code: The block code to be used for decoding.

    Notes:
        - Input type: `hard`.
        - Output type: `hard`.
    """

    code: BlockCode

    def __post_init__(self) -> None:
        self._coset_leaders = self.code.coset_leaders()

    def __call__(self, input: npt.ArrayLike) -> npt.NDArray[np.integer | np.floating]:
        r"""
        Examples:
            >>> code = komm.HammingCode(3)
            >>> decoder = komm.SyndromeTableDecoder(code)
            >>> decoder([[1, 1, 0, 1, 0, 1, 1], [1, 0, 1, 1, 0, 0, 0]])
            array([[1, 1, 0, 0],
                   [1, 0, 1, 1]])
        """

        @blockwise(self.code.length)
        def decode(r: npt.NDArray[np.integer]):
            s = self.code.check(r)
            e_hat = self._coset_leaders[bits_to_int(s)]
            v_hat = np.bitwise_xor(r, e_hat)
            u_hat = self.code.project_word(v_hat)
            return u_hat

        return decode(input)



from .LloydMaxQuantizer import LloydMaxQuantizer
from .ScalarQuantizer import ScalarQuantizer
from .UniformQuantizer import UniformQuantizer

__all__ = [
    "LloydMaxQuantizer",
    "ScalarQuantizer",
    "UniformQuantizer",
]



from functools import cache, cached_property
from typing import SupportsInt

import numpy as np
import numpy.typing as npt

from .._algebra import BinaryPolynomial
from .._util.decorators import blockwise, vectorize
from . import base


class CyclicCode(base.BlockCode):
    r"""
    General binary cyclic code. A cyclic code is a [linear block code](/ref/BlockCode) such that, if $c$ is a codeword, then every cyclic shift of $c$ is also a codeword. It is characterized by its *generator polynomial* $g(X)$, of degree $m$ (the redundancy of the code), and by its *check polynomial* $h(X)$, of degree $k$ (the dimension of the code). Those polynomials are related by $g(X) h(X) = X^n + 1$, where $n = k + m$ is the length of the code.

    Examples of generator polynomials can be found in the table below.

    | Code $(n, k, d)$  | Generator polynomial $g(X)$              | Integer representation           |
    | ----------------- | ---------------------------------------- | -------------------------------- |
    | Hamming $(7,4,3)$ | $X^3 + X + 1$                            | `0b1011 = 0o13 = 11`             |
    | Simplex $(7,3,4)$ | $X^4 + X^2 + X +   1$                    | `0b10111 = 0o27 = 23`            |
    | BCH $(15,5,7)$    | $X^{10} + X^8 + X^5 + X^4 + X^2 + X + 1$ | `0b10100110111 = 0o2467 = 1335`  |
    | Golay $(23,12,7)$ | $X^{11} + X^9 + X^7 + X^6 + X^5 + X + 1$ | `0b101011100011 = 0o5343 = 2787` |

    For more details, see <cite>LC04, Ch. 5</cite>.

    The constructor expects either the generator polynomial or the check polynomial.

    Attributes:
        length: The length $n$ of the code.

        generator_polynomial: The generator polynomial $g(X)$ of the code, of degree $m$ (the redundancy of the code), specified either as a [binary polynomial](/ref/BinaryPolynomial) or as an integer to be converted to the former.

        check_polynomial: The check polynomial $h(X)$ of the code, of degree $k$ (the dimension of the code), specified either as a [binary polynomial](/ref/BinaryPolynomial) or as an integer to be converted to the former.

        systematic: Whether the encoder is systematic. Default is `True`.

    Examples:
        >>> code = komm.CyclicCode(length=23, generator_polynomial=0b101011100011)  # Golay (23, 12)
        >>> (code.length, code.dimension, code.redundancy)
        (23, 12, 11)
        >>> code.minimum_distance()
        7

        >>> code = komm.CyclicCode(length=23, check_polynomial=0b1010010011111)  # Golay (23, 12)
        >>> (code.length, code.dimension, code.redundancy)
        (23, 12, 11)
        >>> code.minimum_distance()
        7
    """

    def __init__(
        self,
        length: int,
        generator_polynomial: SupportsInt | None = None,
        check_polynomial: SupportsInt | None = None,
        systematic: bool = True,
    ):
        self._length = length
        if generator_polynomial is None and check_polynomial is None:
            raise ValueError(
                "either 'generator_polynomial' or 'check_polynomial' must be provided"
            )
        if generator_polynomial is not None and check_polynomial is None:
            self._generator_polynomial = BinaryPolynomial(generator_polynomial)
            modulus, remainder = divmod(self.modulus, self.generator_polynomial)
            if remainder != 0b0:
                raise ValueError("'generator_polynomial' must be a factor of X^n + 1")
            self._check_polynomial = modulus
        elif generator_polynomial is None and check_polynomial is not None:
            self._check_polynomial = BinaryPolynomial(check_polynomial)
            modulus, remainder = divmod(self.modulus, self.check_polynomial)
            if remainder != 0b0:
                raise ValueError("'check_polynomial' must be a factor of X^n + 1")
            self._generator_polynomial = modulus
        self.systematic = systematic

    def __repr__(self) -> str:
        args = f"length={self.length}"
        if self._generator_polynomial is not None:
            args += f", generator_polynomial={self.generator_polynomial}"
        if self._check_polynomial is not None:
            args += f", check_polynomial={self.check_polynomial}"
        args += f", systematic={self.systematic}"
        return f"{self.__class__.__name__}({args})"

    @cached_property
    def length(self) -> int:
        return self._length

    @cached_property
    def dimension(self) -> int:
        return self.check_polynomial.degree

    @cached_property
    def redundancy(self) -> int:
        return self.generator_polynomial.degree

    @cached_property
    def rate(self) -> float:
        return super().rate

    @cached_property
    def generator_polynomial(self) -> BinaryPolynomial:
        return self._generator_polynomial

    @cached_property
    def check_polynomial(self) -> BinaryPolynomial:
        return self._check_polynomial

    @cached_property
    def modulus(self) -> BinaryPolynomial:
        return BinaryPolynomial.from_exponents([0, self.length])

    @cached_property
    def generator_matrix(self) -> npt.NDArray[np.integer]:
        # See [McE04, Sec. 8.1].
        n, k, m = self.length, self.dimension, self.redundancy
        generator_matrix = np.empty((k, n), dtype=int)
        X = BinaryPolynomial(0b10)  # The polynomial X
        if not self.systematic:
            for i in range(k):
                row_poly = X**i * self.generator_polynomial
                generator_matrix[i] = row_poly.coefficients(width=n)
        else:
            for i in range(k):
                row_poly = X ** (m + i) + X ** (m + i) % self.generator_polynomial
                generator_matrix[i] = row_poly.coefficients(width=n)
        return generator_matrix

    @cached_property
    def generator_matrix_right_inverse(self) -> npt.NDArray[np.integer]:
        raise NotImplementedError

    @cached_property
    def check_matrix(self) -> npt.NDArray[np.integer]:
        # See [McE04, Sec. 8.1].
        n, m = self.length, self.redundancy
        check_matrix = np.empty((m, n), dtype=int)
        X = BinaryPolynomial(0b10)  # The polynomial X
        if not self.systematic:
            for i in range(m):
                row_poly = X**i * self.check_polynomial.reciprocal()
                check_matrix[i] = row_poly.coefficients(width=n)
        else:
            for j in range(n):
                col_poly = X**j % self.generator_polynomial
                check_matrix[:, j] = col_poly.coefficients(width=m)
        return check_matrix

    def encode(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        @blockwise(self.dimension)
        @vectorize
        def encode(u: npt.NDArray[np.integer]) -> npt.NDArray[np.integer]:
            u_poly = BinaryPolynomial.from_coefficients(u)
            if not self.systematic:
                v_poly = u_poly * self.generator_polynomial
            else:
                u_poly_shifted = u_poly << self.redundancy
                b_poly = u_poly_shifted % self.generator_polynomial
                v_poly = u_poly_shifted + b_poly
            v = v_poly.coefficients(width=self.length)
            return v

        return encode(input)

    def project_word(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        @blockwise(self.length)
        @vectorize
        def project(v: npt.NDArray[np.integer]) -> npt.NDArray[np.integer]:
            v_poly = BinaryPolynomial.from_coefficients(v)
            if not self.systematic:
                u_poly = v_poly // self.generator_polynomial
            else:
                u_poly = v_poly >> self.redundancy
            u = u_poly.coefficients(width=self.dimension)
            return u

        return project(input)

    def inverse_encode(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        return super().inverse_encode(input)

    def check(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        @blockwise(self.length)
        @vectorize
        def check(r: npt.NDArray[np.integer]) -> npt.NDArray[np.integer]:
            r_poly = BinaryPolynomial.from_coefficients(r)
            s_poly = r_poly % self.generator_polynomial
            s = s_poly.coefficients(width=self.redundancy)
            return s

        return check(input)

    @cache
    def codewords(self) -> npt.NDArray[np.integer]:
        return super().codewords()

    @cache
    def codeword_weight_distribution(self) -> npt.NDArray[np.integer]:
        return super().codeword_weight_distribution()

    @cache
    def minimum_distance(self) -> int:
        return super().minimum_distance()

    @cache
    def coset_leaders(self) -> npt.NDArray[np.integer]:
        return super().coset_leaders()

    @cache
    def coset_leader_weight_distribution(self) -> npt.NDArray[np.integer]:
        return super().coset_leader_weight_distribution()

    @cache
    def packing_radius(self) -> int:
        return super().packing_radius()

    @cache
    def covering_radius(self) -> int:
        return super().covering_radius()



from functools import cached_property
from typing import TypeVar

import numpy as np
import numpy.typing as npt

from . import base

T = TypeVar("T", np.floating, np.complexfloating)


class Modulation(base.Modulation[T]):
    r"""
    General modulation scheme. A *modulation scheme* of *order* $M = 2^m$ is defined by a *constellation* $\mathbf{X}$, which is a real or complex vector of length $M$, and a *binary labeling* $\mathbf{Q}$, which is an $M \times m$ binary matrix whose rows are all distinct. The $i$-th element of $\mathbf{X}$, for $i \in [0:M)$, is denoted by $x_i$ and is called the $i$-th *constellation symbol*. The $i$-th row of $\mathbf{Q}$, for $i \in [0:M)$, is called the *binary representation* of the $i$-th constellation symbol. For more details, see <cite>SA15, Sec. 2.5</cite>."

    Parameters:
        constellation: The constellation $\mathbf{X}$ of the modulation. Must be a 1D-array containing $M$ real or complex numbers.

        labeling: The binary labeling $\mathbf{Q}$ of the modulation. Must be a 2D-array of shape $(M, m)$ where each row is a distinct binary $m$-tuple.

    Examples:
        1. The real modulation scheme depicted in the figure below has $M = 4$ and $m = 2$.
            <figure markdown>
            ![Example for real modulation with M = 4](/figures/modulation_real_4.svg)
            </figure>
            The constellation and labeling are given, respectively, by
            $$
            \mathbf{X} = \begin{bmatrix}
                -0.5 \\\\
                0.0 \\\\
                0.5 \\\\
                2.0
            \end{bmatrix}
            \qquad\text{and}\qquad
            \mathbf{Q} = \begin{bmatrix}
                1 & 0 \\\\
                1 & 1 \\\\
                0 & 1 \\\\
                0 & 0
            \end{bmatrix}.
            $$

                >>> modulation = komm.Modulation(
                ...     constellation=[-0.5, 0.0, 0.5, 2.0],
                ...     labeling=[[1, 0], [1, 1], [0, 1], [0, 0]],
                ... )

        1. The complex modulation scheme depicted in the figure below has $M = 4$ and $m = 2$.
            <figure markdown>
            ![Example for complex modulation with M = 4](/figures/modulation_complex_4.svg)
            </figure>
            The constellation and labeling are given, respectively, by
            $$
            \mathbf{X} = \begin{bmatrix}
                0 \\\\
                -1 \\\\
                1 \\\\
                \mathrm{j}
            \end{bmatrix}
            \qquad\text{and}\qquad
            \mathbf{Q} = \begin{bmatrix}
                0 & 0 \\\\
                0 & 1 \\\\
                1 & 0 \\\\
                1 & 1
            \end{bmatrix}.
            $$

                >>> modulation = komm.Modulation(
                ...     constellation=[0, -1, 1, 1j],
                ...     labeling=[[0, 0], [0, 1], [1, 0], [1, 1]],
                ... )
    """

    def __init__(self, constellation: npt.ArrayLike, labeling: npt.ArrayLike) -> None:
        self._constellation = np.asarray(constellation)
        self._labeling = np.asarray(labeling)
        super()._validate_parameters()

    def __repr__(self) -> str:
        args = ", ".join([
            f"constellation={self.constellation.tolist()}",
            f"labeling={self.labeling.tolist()}",
        ])
        return f"{self.__class__.__name__}({args})"

    @cached_property
    def constellation(self) -> npt.NDArray[T]:
        r"""
        Examples:
            >>> modulation = komm.Modulation(
            ...     constellation=[-0.5, 0.0, 0.5, 2.0],
            ...     labeling=[[1, 0], [1, 1], [0, 1], [0, 0]],
            ... )
            >>> modulation.constellation
            array([-0.5,  0. ,  0.5,  2. ])
        """
        return self._constellation

    @cached_property
    def labeling(self) -> npt.NDArray[np.integer]:
        r"""
        Examples:
            >>> modulation = komm.Modulation(
            ...     constellation=[-0.5, 0.0, 0.5, 2.0],
            ...     labeling=[[1, 0], [1, 1], [0, 1], [0, 0]],
            ... )
            >>> modulation.labeling
            array([[1, 0],
                   [1, 1],
                   [0, 1],
                   [0, 0]])
        """
        return self._labeling

    @cached_property
    def inverse_labeling(self) -> dict[tuple[int, ...], int]:
        r"""
        Examples:
            >>> modulation = komm.Modulation(
            ...     constellation=[-0.5, 0.0, 0.5, 2.0],
            ...     labeling=[[1, 0], [1, 1], [0, 1], [0, 0]],
            ... )
            >>> modulation.inverse_labeling
            {(1, 0): 0, (1, 1): 1, (0, 1): 2, (0, 0): 3}
        """
        return super().inverse_labeling

    @cached_property
    def order(self) -> int:
        r"""
        Examples:
            >>> modulation = komm.Modulation(
            ...     constellation=[-0.5, 0.0, 0.5, 2.0],
            ...     labeling=[[1, 0], [1, 1], [0, 1], [0, 0]],
            ... )
            >>> modulation.order
            4
        """
        return super().order

    @cached_property
    def bits_per_symbol(self) -> int:
        r"""
        Examples:
            >>> modulation = komm.Modulation(
            ...     constellation=[-0.5, 0.0, 0.5, 2.0],
            ...     labeling=[[1, 0], [1, 1], [0, 1], [0, 0]],
            ... )
            >>> modulation.bits_per_symbol
            2
        """
        return super().bits_per_symbol

    @cached_property
    def energy_per_symbol(self) -> float:
        r"""
        Examples:
            >>> modulation = komm.Modulation(
            ...     constellation=[-0.5, 0.0, 0.5, 2.0],
            ...     labeling=[[1, 0], [1, 1], [0, 1], [0, 0]],
            ... )
            >>> modulation.energy_per_symbol
            1.125

            >>> modulation = komm.Modulation(
            ...     constellation=[0, -1, 1, 1j],
            ...     labeling=[[0, 0], [0, 1], [1, 0], [1, 1]],
            ... )
            >>> modulation.energy_per_symbol
            0.75
        """
        return super().energy_per_symbol

    @cached_property
    def energy_per_bit(self) -> float:
        r"""
        Examples:
            >>> modulation = komm.Modulation(
            ...     constellation=[-0.5, 0.0, 0.5, 2.0],
            ...     labeling=[[1, 0], [1, 1], [0, 1], [0, 0]],
            ... )
            >>> modulation.energy_per_bit
            0.5625

            >>> modulation = komm.Modulation(
            ...     constellation=[0, -1, 1, 1j],
            ...     labeling=[[0, 0], [0, 1], [1, 0], [1, 1]],
            ... )
            >>> modulation.energy_per_bit
            0.375
        """
        return super().energy_per_bit

    @cached_property
    def symbol_mean(self) -> float | complex:
        r"""
        Examples:
            >>> modulation = komm.Modulation(
            ...     constellation=[-0.5, 0.0, 0.5, 2.0],
            ...     labeling=[[1, 0], [1, 1], [0, 1], [0, 0]],
            ... )
            >>> modulation.symbol_mean
            0.5

            >>> modulation = komm.Modulation(
            ...     constellation=[0, -1, 1, 1j],
            ...     labeling=[[0, 0], [0, 1], [1, 0], [1, 1]],
            ... )
            >>> modulation.symbol_mean
            0.25j
        """
        return super().symbol_mean

    @cached_property
    def minimum_distance(self) -> float:
        r"""
        Examples:
            >>> modulation = komm.Modulation(
            ...     constellation=[-0.5, 0.0, 0.5, 2.0],
            ...     labeling=[[1, 0], [1, 1], [0, 1], [0, 0]],
            ... )
            >>> modulation.minimum_distance
            0.5

            >>> modulation = komm.Modulation(
            ...     constellation=[0, -1, 1, 1j],
            ...     labeling=[[0, 0], [0, 1], [1, 0], [1, 1]],
            ... )
            >>> modulation.minimum_distance
            1.0
        """
        return super().minimum_distance

    def modulate(self, input: npt.ArrayLike) -> npt.NDArray[T]:
        r"""
        Examples:
            >>> modulation = komm.Modulation(
            ...     constellation=[-0.5, 0.0, 0.5, 2.0],
            ...     labeling=[[1, 0], [1, 1], [0, 1], [0, 0]],
            ... )
            >>> modulation.modulate([0, 0, 1, 1, 0, 0, 1, 0])
            array([ 2. ,  0. ,  2. , -0.5])
            >>> modulation.modulate([[0, 0, 1, 1], [0, 0, 1, 0]])
            array([[ 2. ,  0. ],
                   [ 2. , -0.5]])
        """
        return super().modulate(input)

    def demodulate_hard(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Note:
            This method implements the general minimum Euclidean distance hard demodulator, assuming uniformly distributed symbols.

        Examples:
            >>> modulation = komm.Modulation(
            ...     constellation=[-0.5, 0.0, 0.5, 2.0],
            ...     labeling=[[1, 0], [1, 1], [0, 1], [0, 0]],
            ... )
            >>> modulation.demodulate_hard([2.17, -0.06, 1.94, -0.61])
            array([0, 0, 1, 1, 0, 0, 1, 0])
            >>> modulation.demodulate_hard([[2.17, -0.06], [1.94, -0.61]])
            array([[0, 0, 1, 1],
                   [0, 0, 1, 0]])
        """
        return super().demodulate_hard(input)

    def demodulate_soft(
        self, input: npt.ArrayLike, snr: float = 1.0
    ) -> npt.NDArray[np.floating]:
        r"""
        Note:
            This method implements the general soft demodulator, assuming uniformly distributed symbols.

        Examples:
            >>> modulation = komm.Modulation(
            ...     constellation=[-0.5, 0.0, 0.5, 2.0],
            ...     labeling=[[1, 0], [1, 1], [0, 1], [0, 0]],
            ... )
            >>> modulation.demodulate_soft([2.17, -0.06, 1.94, -0.61], snr=100.0).round(1)
            array([ 416. ,  245.3,  -27.6,  -16.9,  334.2,  184. , -108.4,   32. ])
            >>> modulation.demodulate_soft([[2.17, -0.06], [1.94, -0.61]], snr=100.0).round(1)
            array([[ 416. ,  245.3,  -27.6,  -16.9],
                   [ 334.2,  184. , -108.4,   32. ]])
        """
        return super().demodulate_soft(input, snr)



from collections.abc import Callable
from functools import wraps
from typing import Any, TypeVar

import numpy as np
import numpy.typing as npt

T = TypeVar("T", bound=np.generic)
U = TypeVar("U", bound=np.generic)


ArrayFunction = Callable[[npt.NDArray[T]], npt.NDArray[U]]


def vectorize(func: ArrayFunction[T, U]):
    r"""
    Vectorizes a function that accepts a 1D array and returns a 1D array.
    """

    @wraps(func)
    def wrapper(arr: npt.NDArray[T], *args: Any, **kwargs: Any):
        return np.apply_along_axis(func, -1, arr, *args, **kwargs)

    return wrapper


def blockwise(block_size: int):
    r"""
    Applies a function blockwise to the last dimension of an array.
    """

    def decorator(func: ArrayFunction[T, U]):
        @wraps(func)
        def wrapper(arr: npt.ArrayLike):
            arr = np.asarray(arr)
            if arr.shape[-1] % block_size != 0:
                raise ValueError(
                    "last dimension of array must be a multiple of block size"
                    f" {block_size} (got {arr.shape[-1]})"
                )
            blocks = arr.reshape(*arr.shape[:-1], -1, block_size)
            processed = func(blocks)
            return processed.reshape(*processed.shape[:-2], -1)

        return wrapper

    return decorator



from .bit_operations import bits_to_int, int_to_bits
from .correlation import autocorrelation, cyclic_autocorrelation
from .information_theory import binary_entropy, entropy, relative_entropy
from .special_functions import gaussian_q, gaussian_q_inv, marcum_q

__all__ = [
    "bits_to_int",
    "int_to_bits",
    "autocorrelation",
    "cyclic_autocorrelation",
    "entropy",
    "binary_entropy",
    "relative_entropy",
    "gaussian_q",
    "gaussian_q_inv",
    "marcum_q",
]



from .ConvolutionalCode import ConvolutionalCode
from .ConvolutionalStreamEncoder import ConvolutionalStreamEncoder
from .TerminatedConvolutionalCode import TerminatedConvolutionalCode
from .ViterbiStreamDecoder import ViterbiStreamDecoder

__all__ = [
    "ConvolutionalCode",
    "TerminatedConvolutionalCode",
    "ConvolutionalStreamEncoder",
    "ViterbiStreamDecoder",
]



from dataclasses import dataclass
from functools import cache
from heapq import heapify, heappop, heappush
from itertools import product
from math import ceil, log2

import numpy.typing as npt
from tqdm import tqdm
from typing_extensions import Self

from .._util.docs import mkdocstrings
from .._util.information_theory import PMF
from .util import Word
from .VariableToFixedCode import VariableToFixedCode


@mkdocstrings(filters=["!.*"])
class TunstallCode(VariableToFixedCode):
    r"""
    Binary Tunstall code. It is an optimal (minimal expected rate) [variable-to-fixed length code](/ref/VariableToFixedCode) for a given probability mass function. For more details, see <cite>Say06, Sec. 3.7</cite>.

    Notes:
        Tunstall codes are always [prefix-free](/ref/VariableToFixedCode/#is_prefix_free) (hence [uniquely encodable](/ref/VariableToFixedCode/#is_uniquely_encodable)) and [fully covering](/ref/VariableToFixedCode/#is_fully_covering).

    Parameters:
        pmf: The probability mass function of the source.
        target_block_size: The target block size $n$. Must satisfy $2^n \geq S$, where $S$ is the cardinality of the source alphabet, given by `len(pmf)`. The default value is $n = \lceil \log_2 S \rceil$.

    Examples:
        >>> pmf = [0.7, 0.15, 0.15]

        >>> code = komm.TunstallCode(pmf)
        >>> code.dec_mapping
        {(0, 0): (0,),
         (0, 1): (1,),
         (1, 0): (2,)}
        >>> code.rate(pmf)  # doctest: +FLOAT_CMP
        np.float64(2.0)

        >>> code = komm.TunstallCode(pmf, 3)
        >>> code.dec_mapping
        {(0, 0, 0): (0, 0, 0),
         (0, 0, 1): (0, 0, 1),
         (0, 1, 0): (0, 0, 2),
         (0, 1, 1): (0, 1),
         (1, 0, 0): (0, 2),
         (1, 0, 1): (1,),
         (1, 1, 0): (2,)}
        >>> code.rate(pmf)  # doctest: +FLOAT_CMP
        np.float64(1.3698630137)
    """

    def __init__(
        self,
        pmf: npt.ArrayLike,
        target_block_size: int | None = None,
    ) -> None:
        self.pmf = PMF(pmf)
        if target_block_size is None:
            target_block_size = ceil(log2(self.pmf.size))
        if 2**target_block_size < self.pmf.size:
            raise ValueError("'target_block_size' is too low")
        super().__init__(
            target_cardinality=2,
            source_cardinality=self.pmf.size,
            target_block_size=target_block_size,
            dec_mapping=tunstall_algorithm(self.pmf, target_block_size),
        )

    def __repr__(self) -> str:
        args = ", ".join([
            f"pmf={self.pmf.tolist()}",
            f"target_block_size={self.target_block_size}",
        ])
        return f"{self.__class__.__name__}({args})"

    @cache
    def is_fully_covering(self) -> bool:
        return True

    @cache
    def is_uniquely_encodable(self) -> bool:
        return True

    @cache
    def is_prefix_free(self) -> bool:
        return True


def tunstall_algorithm(pmf: PMF, code_block_size: int) -> dict[Word, Word]:
    @dataclass
    class Node:
        sourceword: Word
        probability: float

        def __lt__(self, other: Self) -> bool:
            return -self.probability < -other.probability

    pbar = tqdm(
        desc="Generating Tunstall code",
        total=2 ** (code_block_size - 1) - pmf.size + 1,
        delay=2.5,
    )

    heap = [Node((symbol,), probability) for (symbol, probability) in enumerate(pmf)]
    heapify(heap)
    while len(heap) + pmf.size - 1 < 2**code_block_size:
        node = heappop(heap)
        for symbol, probability in enumerate(pmf):
            new_node = Node(node.sourceword + (symbol,), node.probability * probability)
            heappush(heap, new_node)
        pbar.update()

    pbar.close()

    dec_mapping = dict(
        zip(
            product([0, 1], repeat=code_block_size),
            sorted(node.sourceword for node in heap),
        )
    )
    return dec_mapping



from abc import ABC, abstractmethod
from dataclasses import dataclass
from functools import cached_property

import numpy as np
import numpy.typing as npt
from numpy.linalg import matrix_power

from .._error_control_convolutional import TerminatedConvolutionalCode
from .._util.bit_operations import bits_to_int, int_to_bits
from .._util.matrices import pseudo_inverse
from .ConvolutionalCode import ConvolutionalCode


@dataclass
class TerminationStrategy(ABC):
    convolutional_code: "ConvolutionalCode"
    num_blocks: int

    @abstractmethod
    def initial_state(self, input_sequence: npt.ArrayLike) -> int: ...

    @abstractmethod
    def pre_process_input(
        self, input_bits: npt.ArrayLike
    ) -> npt.NDArray[np.integer]: ...

    @abstractmethod
    def codeword_length(self) -> int: ...

    @abstractmethod
    def generator_matrix(
        self, code: TerminatedConvolutionalCode
    ) -> npt.NDArray[np.integer]: ...


def _base_generator_matrix(
    code: TerminatedConvolutionalCode,
    convolutional_code: ConvolutionalCode,
    num_blocks: int,
) -> npt.NDArray[np.integer]:
    k0 = convolutional_code.num_input_bits
    n0 = convolutional_code.num_output_bits
    k, n = code.dimension, code.length
    generator_matrix = np.zeros((k, n), dtype=int)
    top_rows = np.apply_along_axis(code.encode, 1, np.eye(k0, k, dtype=int))
    for t in range(num_blocks):
        generator_matrix[k0 * t : k0 * (t + 1), :] = np.roll(top_rows, n0 * t, 1)
    return generator_matrix


@dataclass
class DirectTruncation(TerminationStrategy):
    def initial_state(self, input_sequence: npt.ArrayLike) -> int:
        return 0

    def pre_process_input(self, input_bits: npt.ArrayLike) -> npt.NDArray[np.integer]:
        return np.asarray(input_bits)

    def codeword_length(self) -> int:
        h = self.num_blocks
        n0 = self.convolutional_code.num_output_bits
        return h * n0

    def generator_matrix(
        self, code: TerminatedConvolutionalCode
    ) -> npt.NDArray[np.integer]:
        h = self.num_blocks
        k0 = self.convolutional_code.num_input_bits
        n0 = self.convolutional_code.num_output_bits
        generator_matrix = _base_generator_matrix(code, self.convolutional_code, h)
        for t in range(1, h):
            generator_matrix[k0 * t : k0 * (t + 1), : n0 * t] = 0
        return generator_matrix


@dataclass
class ZeroTermination(TerminationStrategy):
    def initial_state(self, input_sequence: npt.ArrayLike) -> int:
        return 0

    def pre_process_input(self, input_bits: npt.ArrayLike) -> npt.NDArray[np.integer]:
        input_bits = np.asarray(input_bits)
        tail = input_bits @ self._tail_projector % 2
        return np.concatenate([input_bits, tail])

    def codeword_length(self) -> int:
        h = self.num_blocks
        n0 = self.convolutional_code.num_output_bits
        m = self.convolutional_code.memory_order
        return (h + m) * n0

    def generator_matrix(
        self, code: TerminatedConvolutionalCode
    ) -> npt.NDArray[np.integer]:
        return _base_generator_matrix(code, self.convolutional_code, self.num_blocks)

    @cached_property
    def _tail_projector(self) -> npt.NDArray[np.integer]:
        h = self.num_blocks
        mu = self.convolutional_code.memory_order
        A_mat, B_mat, _, _ = self.convolutional_code.state_space_representation()
        AnB_message = np.vstack(
            [B_mat @ matrix_power(A_mat, j) % 2 for j in range(mu + h - 1, mu - 1, -1)]
        )
        AnB_tail = np.vstack(
            [B_mat @ matrix_power(A_mat, j) % 2 for j in range(mu - 1, -1, -1)]
        )
        return AnB_message @ pseudo_inverse(AnB_tail) % 2


@dataclass
class TailBiting(TerminationStrategy):
    def initial_state(self, input_sequence: npt.ArrayLike) -> int:
        fsm = self.convolutional_code.finite_state_machine()
        nu = self.convolutional_code.overall_constraint_length
        _, zs_response = fsm.process(input_sequence, initial_state=0)
        zs_response = int_to_bits(zs_response, width=nu)
        initial_state = bits_to_int(zs_response @ self._zs_multiplier % 2)
        assert isinstance(initial_state, int)
        return initial_state

    def pre_process_input(self, input_bits: npt.ArrayLike) -> npt.NDArray[np.integer]:
        return np.asarray(input_bits)

    def codeword_length(self) -> int:
        h = self.num_blocks
        n0 = self.convolutional_code.num_output_bits
        return h * n0

    def generator_matrix(
        self, code: TerminatedConvolutionalCode
    ) -> npt.NDArray[np.integer]:
        return _base_generator_matrix(code, self.convolutional_code, self.num_blocks)

    @cached_property
    def _zs_multiplier(self) -> npt.NDArray[np.integer]:
        h = self.num_blocks
        nu = self.convolutional_code.overall_constraint_length
        A_mat, _, _, _ = self.convolutional_code.state_space_representation()
        return pseudo_inverse(matrix_power(A_mat, h) + np.eye(nu, dtype=int) % 2)



from dataclasses import dataclass, field
from itertools import product
from math import prod
from typing import Any

import numpy as np
import numpy.typing as npt

from .._util.information_theory import PMF

Word = tuple[int, ...]


def is_prefix_of(w1: Word, w2: Word) -> bool:
    return len(w1) <= len(w2) and w2[: len(w1)] == w1


def is_prefix_free(words: list[Word]) -> bool:
    words = sorted(words, key=len)
    for i, w1 in enumerate(words):
        for w2 in words[i + 1 :]:
            if is_prefix_of(w1, w2):
                return False
    return True


def is_uniquely_parsable(words: list[Word]) -> bool:
    # Sardinas–Patterson algorithm. See [Say06, Sec. 2.4.1].
    augmented_words = set(words)
    while True:
        dangling_suffixes: set[Word] = set()
        for w1, w2 in product(set(words), augmented_words):
            if w1 == w2:
                continue
            if is_prefix_of(w1, w2):
                dangling_suffixes.add(w2[len(w1) :])
            elif is_prefix_of(w2, w1):
                dangling_suffixes.add(w1[len(w2) :])
        if dangling_suffixes & set(words):
            return False
        if dangling_suffixes <= augmented_words:
            return True
        augmented_words |= dangling_suffixes


def is_fully_covering(words: list[Word], cardinality: int) -> bool:
    @dataclass(eq=False)
    class Node:
        is_end: bool = False
        children: dict[int, "Node"] = field(default_factory=dict)

    # Build trie
    root = Node()
    for word in words:
        node = root
        for symbol in word:
            if symbol not in node.children:
                node.children[symbol] = Node()
            node = node.children[symbol]
        node.is_end = True

    visited = {root}
    stack = [root]
    while stack:
        node = stack.pop()
        for symbol in range(cardinality):
            if symbol not in node.children:
                return False
            child = node.children[symbol]
            if not child.is_end and child not in visited:
                visited.add(child)
                stack.append(child)

    return True


def parse_fixed_length(
    input: npt.NDArray[np.integer],
    dictionary: dict[Word, Word],
    block_size: int,
) -> npt.NDArray[np.integer]:
    if input.size % block_size != 0:
        raise ValueError(
            "length of input must be a multiple of block size"
            f" {block_size} (got {len(input)})"
        )
    try:
        output = np.concatenate(
            [dictionary[tuple(w)] for w in input.reshape(-1, block_size)]
        )
    except KeyError:
        raise ValueError("input contains invalid word")
    return output


def parse_prefix_free(
    input: npt.NDArray[np.integer],
    dictionary: dict[Word, Word],
    allow_incomplete: bool,
) -> npt.NDArray[np.integer]:
    output: list[int] = []
    i = 0
    for j in range(len(input)):
        key = tuple(input[i : j + 1])
        if key in dictionary:
            output.extend(dictionary[key])
            i = j + 1

    if i == len(input):
        return np.asarray(output)
    elif not allow_incomplete:
        raise ValueError("input contains invalid word")

    remainder = tuple(input[i:])
    for key, value in dictionary.items():
        if is_prefix_of(remainder, key):
            output.extend(value)
            return np.asarray(output)

    raise ValueError("input contains invalid word")


def extended_probabilities(pmf: PMF, k: int, pbar: Any) -> list[tuple[Word, float]]:
    probs: list[tuple[float, Word]] = []
    for u in product(range(pmf.size), repeat=k):
        pu = prod(pmf[list(u)])
        probs.append((-pu, u))
        if pbar is not None:
            pbar.update()
    return [(u, -minus_pu) for minus_pu, u in sorted(probs)]


def empty_mapping(cardinality: int, block_size: int) -> dict[Word, Word]:
    return {x: () for x in product(range(cardinality), repeat=block_size)}


def integer_to_symbols(integer: int, base: int, width: int) -> list[int]:
    symbols: list[int] = []
    for _ in range(width):
        integer, symbol = divmod(integer, base)
        symbols.append(symbol)
    return symbols[::-1]


def symbols_to_integer(symbols: npt.ArrayLike, base: int) -> int:
    symbols = np.asarray(symbols)
    integer = 0
    for symbol in symbols:
        integer = integer * base + symbol
    return integer



from math import gamma
from statistics import NormalDist

import numpy as np
import numpy.typing as npt

norm = NormalDist()

gamma = np.vectorize(gamma)


def _gaussian_q(x: float) -> float:
    return norm.cdf(-x)


def gaussian_q(x: npt.ArrayLike) -> npt.NDArray[np.floating] | np.floating:
    r"""
    Computes the Gaussian Q-function. It is given by
    $$
        \mathrm{Q}(x) = \frac{1}{\sqrt{2\pi}} \int_x^\infty \mathrm{e}^{-u^2/2} \, \mathrm{d}u.
    $$
    This corresponds to the complementary cumulative distribution function of the standard gaussian distribution. For more details, see [Wikipedia: Q-function](https://en.wikipedia.org/wiki/Q-function).

    Parameters:
        x: The input to the function. Should be a float or array of floats.

    Returns:
        y: The value $y = \mathrm{Q}(x)$.

    Examples:
        >>> komm.gaussian_q(0.0)
        np.float64(0.5)

        >>> komm.gaussian_q([[-1.0], [0.0], [1.0]])
        array([[0.84134475],
               [0.5       ],
               [0.15865525]])
    """
    result = np.vectorize(_gaussian_q)(x)
    return result[()] if np.ndim(result) == 0 else result


def _gaussian_q_inv(y: float) -> float:
    if y == 0:
        return np.inf
    if y == 1:
        return -np.inf
    return -norm.inv_cdf(y)


def gaussian_q_inv(y: npt.ArrayLike) -> npt.NDArray[np.floating] | np.floating:
    r"""
    Computes the inverse Gaussian Q-function.

    Parameters:
        y: The input to the function. Should be a float or array of floats in the real interval $[0, 1]$.

    Returns:
        x: The value $x = \mathrm{Q^{-1}}(y)$.

    Examples:
        >>> komm.gaussian_q_inv(0.5)
        np.float64(0.0)

        >>> komm.gaussian_q_inv([[0.841344746], [0.5], [0.158655254]])
        array([[-1.],
               [ 0.],
               [ 1.]])
    """
    result = np.vectorize(_gaussian_q_inv)(y) + 0.0  # + 0.0 to avoid -0.0
    return result[()] if np.ndim(result) == 0 else result


def _auxiliary_p(
    m: int, a: float, b: float, max_iter: int = 100, tol: float = 1e-16
) -> float:
    # Auxiliary function for the Marcum Q-function
    # See [https://en.wikipedia.org/wiki/Marcum_Q-function].
    ms = np.arange(m)
    result = np.sum(np.exp(-b) * (b**ms) / gamma(ms + 1))
    for i in range(m, max_iter):
        ks = np.arange(i - m + 1)
        inner_sum = np.sum(np.exp(-m * a) * ((m * a) ** ks) / gamma(ks + 1))
        term = np.exp(-b) * (b**i) / gamma(i + 1) * (1 - inner_sum)
        result += term
        if abs(term) < tol:
            break
    return result


def marcum_q(
    m: int, a: npt.ArrayLike, x: npt.ArrayLike
) -> npt.NDArray[np.floating] | np.floating:
    r"""
    Computes the Marcum Q-function. It is given by
    $$
        \mathrm{Q}_m(a; x) = \int_x^\infty u \left( \frac{u}{a} \right)^{m-1}  I\_{m-1}(a x) \exp \left( -\frac{u^2 + a^2}{2} \right) \mathrm{d}u,
    $$
    where $I\_{m-1}$ is the modified Bessel function of the first kind. This corresponds to the complementary cumulative distribution function of the non-central chi distribution with $2m$ degrees of freedom and non-centrality parameter $a$. For more details, see [Wikipedia: Marcum Q-function](https://en.wikipedia.org/wiki/Marcum_Q-function).

    Parameters:
        m: The order of the Marcum Q-function. Should be a positive integer.
        a: The value of $a$. Should be a float or array of floats.
        x: The input to the function. Should be a float or array of floats.

    Returns:
        y: The value $y = \mathrm{Q}_m(a; x)$.

    Examples:
        >>> komm.marcum_q(1, 1, 1)
        np.float64(0.7328798037968204)

        >>> komm.marcum_q(2, 0.5, [1.2, 1.4, 1.6])
        array([0.85225816, 0.76472056, 0.66139663])
    """
    a = np.asarray(a)
    x = np.asarray(x)
    if not m > 0:
        raise ValueError("'m' must be positive")
    if not np.all(a >= 0) or not np.all(x >= 0):
        raise ValueError("'a' and 'x' must be non-negative")
    result = np.vectorize(_auxiliary_p)(m, a**2 / (2 * m), x**2 / 2)
    return result[()] if np.ndim(result) == 0 else result


def logcosh(x: npt.ArrayLike) -> np.floating:
    # https://stackoverflow.com/a/57786270/
    s = np.sign(x) * x
    p = np.exp(-2 * s)
    return s + np.log1p(p) - np.log(2)



from abc import ABC, abstractmethod

import numpy as np
import numpy.typing as npt


class IntegerCode(ABC):
    @abstractmethod
    def encode(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Encode the input integer array.

        Parameters:
            input: The input integer array. It must be a one-dimensional array.

        Returns:
            output: The sequence of bits corresponding to the input integer array.
        """
        raise NotImplementedError

    @abstractmethod
    def decode(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Decode the input bit array.

        Parameters:
            input: The input bit array. It must be a one-dimensional array.

        Returns:
            output: The sequence of integers corresponding to the input bit array.
        """
        raise NotImplementedError



from .AWGNChannel import AWGNChannel
from .BinaryErasureChannel import BinaryErasureChannel
from .BinarySymmetricChannel import BinarySymmetricChannel
from .DiscreteMemorylessChannel import DiscreteMemorylessChannel
from .ZChannel import ZChannel

__all__ = [
    "AWGNChannel",
    "BinaryErasureChannel",
    "BinarySymmetricChannel",
    "DiscreteMemorylessChannel",
    "ZChannel",
]



from abc import ABC, abstractmethod
from functools import cached_property

import numpy as np
import numpy.typing as npt


class Pulse(ABC):
    @abstractmethod
    def waveform(self, t: npt.ArrayLike) -> npt.NDArray[np.floating]:
        raise NotImplementedError

    @abstractmethod
    def spectrum(self, f: npt.ArrayLike) -> npt.NDArray[np.floating]:
        raise NotImplementedError

    @cached_property
    @abstractmethod
    def support(self) -> tuple[float, float]:
        raise NotImplementedError



import numpy as np
import numpy.typing as npt

from .._algebra.BinaryPolynomial import BinaryPolynomial


def barker_sequence(length: int) -> npt.NDArray[np.integer]:
    return np.array(
        {
            2: [0, 1],
            3: [0, 0, 1],
            4: [0, 0, 1, 0],
            5: [0, 0, 0, 1, 0],
            7: [0, 0, 0, 1, 1, 0, 1],
            11: [0, 0, 0, 1, 1, 1, 0, 1, 1, 0, 1],
            13: [0, 0, 0, 0, 0, 1, 1, 0, 0, 1, 0, 1, 0],
        }[length]
    )


def hadamard_matrix(length: int) -> npt.NDArray[np.integer]:
    h = np.array([[1]])
    g = np.array([[1, 1], [1, -1]])
    for _ in range(length.bit_length() - 1):
        h = np.kron(h, g).astype(int)
    return h


def lfsr_sequence(
    feedback_polynomial: BinaryPolynomial,
    start_state_polynomial: BinaryPolynomial,
) -> npt.NDArray[np.integer]:
    taps = (feedback_polynomial + BinaryPolynomial(1)).exponents()
    length = 2 ** taps[-1] - 1
    state = start_state_polynomial.coefficients(width=feedback_polynomial.degree)
    sequence = np.empty(length, dtype=int)
    for i in range(length):
        sequence[i] = state[-1]
        state[-1] = np.count_nonzero(state[taps - 1]) % 2
        state = np.roll(state, 1)
    return sequence



from .APSKModulation import APSKModulation
from .ASKModulation import ASKModulation
from .Modulation import Modulation
from .PAModulation import PAModulation
from .PSKModulation import PSKModulation
from .QAModulation import QAModulation

__all__ = [
    "APSKModulation",
    "ASKModulation",
    "Modulation",
    "PAModulation",
    "PSKModulation",
    "QAModulation",
]



from dataclasses import dataclass
from functools import cache

import numpy as np
import numpy.typing as npt
from tqdm import tqdm

from .._util.bit_operations import int_to_bits
from .._util.docs import mkdocstrings
from .BlockCode import BlockCode


@mkdocstrings(filters=["!.*"])
@dataclass(eq=False)
class Lexicode(BlockCode):
    r"""
    Lexicographic code (lexicode). For a given length $n$ and minimum distance $d$, it is the [linear block code](/ref/BlockCode) obtained by starting with the all-zero codeword and adding all binary $n$-tuples (in lexicographic order) that are at least at distance $d$ from all codewords already in the code.

    Attributes:
        n: The length $n$ of the code.
        d: The minimum distance $d$ of the code.

    For more details, see <cite>HP03, Sec. 2.11</cite>.

    Examples:
        >>> code = komm.Lexicode(7, 3)  # Hamming (7, 4)
        >>> (code.length, code.dimension, code.redundancy)
        (7, 4, 3)
        >>> code.generator_matrix
        array([[0, 0, 0, 0, 1, 1, 1],
               [0, 0, 1, 1, 0, 0, 1],
               [0, 1, 0, 1, 0, 1, 0],
               [1, 0, 0, 1, 0, 1, 1]])
        >>> code.minimum_distance()
        3
    """

    n: int
    d: int

    def __post_init__(self) -> None:
        if not 1 <= self.d <= self.n:
            raise ValueError("'n' and 'd' must satisfy 1 <= d <= n")
        super().__init__(generator_matrix=lexicode_generator_matrix(self.n, self.d))

    @cache
    def minimum_distance(self) -> int:
        return self.d


def lexicode_generator_matrix(n: int, d: int) -> npt.NDArray[np.integer]:
    codewords = [0]
    basis: list[int] = []
    for i in tqdm(range(1, 2**n), desc="Generating lexicode", delay=2.5):
        # Reverse checking is way faster [Wikipedia].
        if all((c ^ i).bit_count() >= d for c in reversed(codewords)):
            if len(codewords).bit_count() == 1:  # Is power of 2.
                basis.append(i)
            codewords.append(i)
    generator_matrix = np.array([list(reversed(int_to_bits(c, n))) for c in basis])
    return generator_matrix



from .BinaryPolynomial import BinaryPolynomial
from .BinaryPolynomialFraction import BinaryPolynomialFraction
from .FiniteBifield import FiniteBifield, FiniteBifieldElement
from .Integers import Integer

__all__ = [
    "BinaryPolynomial",
    "BinaryPolynomialFraction",
    "FiniteBifield",
    "FiniteBifieldElement",
    "Integer",
]



from functools import cached_property

import numpy as np
import numpy.typing as npt

from .._util.information_theory import PMF, LogBase, entropy


class DiscreteMemorylessSource:
    r"""
    Discrete memoryless source (DMS). It is defined by an *alphabet* $\mathcal{X}$ and a *probability mass function* (pmf) $p_X$. Here, for simplicity, the alphabet is always taken as $\mathcal{X} = \\{ 0, 1, \ldots, |\mathcal{X}| - 1 \\}$. The pmf $p_X$ gives the probability of the source emitting the symbol $X = x$.

    Parameters:
        pmf: The source probability mass function $p_X$. The element in position $x \in \mathcal{X}$ must be equal to $p_X(x)$.
    """

    def __init__(
        self, pmf: npt.ArrayLike, rng: np.random.Generator = np.random.default_rng()
    ):
        self.pmf = PMF(pmf)
        self.rng = rng

    def __repr__(self) -> str:
        return f"{__class__.__name__}(pmf={self.pmf.tolist()})"

    @cached_property
    def cardinality(self) -> int:
        r"""
        The cardinality $|\mathcal{X}|$ of the source alphabet.
        """
        return self.pmf.size

    def entropy(self, base: LogBase = 2.0) -> float:
        r"""
        Returns the source entropy $\mathrm{H}(X)$. See [`komm.entropy`](/ref/entropy) for more details.

        Parameters:
            base: See [`komm.entropy`](/ref/entropy). The default value is $2.0$.

        Examples:
            >>> dms = komm.DiscreteMemorylessSource([1/2, 1/4, 1/8, 1/8])
            >>> dms.entropy()
            np.float64(1.75)
            >>> dms.entropy(base=4)
            np.float64(0.875)
        """
        return entropy(self.pmf, base)

    def __call__(self, shape: int | tuple[int, ...] = ()) -> npt.NDArray[np.integer]:
        r"""
        Returns random samples from the source.

        Parameters:
            shape: The shape of the output array. If `shape` is an integer, the output array will have shape `(shape,)`. The default value is `()`, which returns a single sample.

        Returns:
            An array of shape `shape` with random samples from the source.

        Examples:
            >>> rng = np.random.default_rng(seed=42)
            >>> dms = komm.DiscreteMemorylessSource([0.5, 0.4, 0.1], rng=rng)
            >>> dms()
            array(1)
            >>> dms(10)
            array([0, 1, 1, 0, 2, 1, 1, 0, 0, 0])
            >>> dms((2, 5))
            array([[2, 1, 1, 0, 0],
                   [1, 0, 1, 1, 1]])
        """
        return self.rng.choice(self.pmf.size, p=self.pmf, size=shape)



from abc import ABC, abstractmethod
from functools import cached_property
from itertools import combinations
from typing import Generic, TypeVar

import numpy as np
import numpy.typing as npt
from typing_extensions import final

from .._util.decorators import vectorize

T = TypeVar("T", np.floating, np.complexfloating)


class Modulation(ABC, Generic[T]):
    @final
    def _validate_parameters(self) -> None:
        order, m = self.order, self.bits_per_symbol
        if order & (order - 1):
            raise ValueError("length of 'constellation' must be a power of two")
        if self.labeling.shape != (order, m):
            raise ValueError(f"shape of 'labeling' must be ({order}, {m})")
        if np.any(self.labeling < 0) or np.any(self.labeling > 1):
            raise ValueError("elements of 'labeling' must be either 0 or 1")
        if len(set(tuple(row) for row in self.labeling)) != order:
            raise ValueError("rows of 'labeling' must be distinct")

    @cached_property
    @abstractmethod
    def constellation(self) -> npt.NDArray[T]:
        r"""
        The constellation $\mathbf{X}$ of the modulation.
        """
        raise NotImplementedError

    @cached_property
    @abstractmethod
    def labeling(self) -> npt.NDArray[np.integer]:
        r"""
        The labeling $\mathbf{Q}$ of the modulation.
        """
        raise NotImplementedError

    @cached_property
    @abstractmethod
    def inverse_labeling(self) -> dict[tuple[int, ...], int]:
        r"""
        The inverse labeling of the modulation. It is a dictionary that maps each binary tuple to the corresponding constellation index.
        """
        return {tuple(int(x) for x in row): i for i, row in enumerate(self.labeling)}

    @cached_property
    @abstractmethod
    def order(self) -> int:
        r"""
        The order $M$ of the modulation.
        """
        return self.constellation.size

    @cached_property
    @abstractmethod
    def bits_per_symbol(self) -> int:
        r"""
        The number $m$ of bits per symbol of the modulation. It is given by
        $$
            m = \log_2 M,
        $$
        where $M$ is the order of the modulation.
        """
        return (self.order - 1).bit_length()

    @cached_property
    @abstractmethod
    def energy_per_symbol(self) -> float:
        r"""
        The average symbol energy $E_\mathrm{s}$ of the constellation. It assumes equiprobable symbols. It is given by
        $$
          E_\mathrm{s} = \frac{1}{M} \sum_{i \in [0:M)} \lVert x_i \rVert^2,
        $$
        where $\lVert x_i \rVert^2$ is the energy of constellation symbol $x_i$, and $M$ is the order of the modulation.
        """
        return float(np.mean(np.real(self.constellation * self.constellation.conj())))

    @cached_property
    @abstractmethod
    def energy_per_bit(self) -> float:
        r"""
        The average bit energy $E_\mathrm{b}$ of the constellation. It assumes equiprobable symbols. It is given by
        $$
            E_\mathrm{b} = \frac{E_\mathrm{s}}{m},
        $$
        where $E_\mathrm{s}$ is the average symbol energy, and $m$ is the number of bits per symbol of the modulation.
        """
        return self.energy_per_symbol / self.bits_per_symbol

    @cached_property
    @abstractmethod
    def symbol_mean(self) -> float | complex:
        r"""
        The mean $\mu_\mathrm{s}$ of the constellation. It assumes equiprobable symbols. It is given by
        $$
          \mu_\mathrm{s} = \frac{1}{M} \sum_{i \in [0:M)} x_i.
        $$
        """
        mean = self.constellation.mean()
        return complex(mean) if np.iscomplexobj(self.constellation) else float(mean)

    @cached_property
    @abstractmethod
    def minimum_distance(self) -> float:
        r"""
        The minimum Euclidean distance $d_\mathrm{min}$ of the constellation. It is given by
        $$
            d_\mathrm{min} = \min_ { i, j \in [0:M), ~ i \neq j } \lVert x_i - x_j \rVert.
        $$
        """
        return float(
            np.min([np.abs(s1 - s2) for s1, s2 in combinations(self.constellation, 2)])
        )

    @abstractmethod
    def modulate(self, input: npt.ArrayLike) -> npt.NDArray[T]:
        r"""
        Modulates one or more sequences of bits to their corresponding constellation symbols.

        Parameters:
            input: The input sequence(s). Can be either a single sequence whose length is a multiple of $m$, or a multidimensional array where the last dimension is a multiple of $m$.

        Returns:
            output: The output sequence(s). Has the same shape as the input, with the last dimension divided by $m$.
        """

        @vectorize
        def modulate(bits: npt.NDArray[np.integer]) -> npt.NDArray[T]:
            return self.constellation[self.inverse_labeling[tuple(bits)]]

        input = np.asarray(input)
        if input.shape[-1] % self.bits_per_symbol != 0:
            raise ValueError(
                "last dimension of 'bits' must be a multiple of bits per symbol"
                f" {self.bits_per_symbol} (got {input.shape[-1]})"
            )
        bits = input.reshape(*input.shape[:-1], -1, self.bits_per_symbol)
        symbols = modulate(bits)
        output = symbols.reshape(*symbols.shape[:-1], -1)
        return output

    @abstractmethod
    def demodulate_hard(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Demodulates one or more sequences of received points to their corresponding sequences of hard bits ($\mathtt{0}$ or $\mathtt{1}$) using hard-decision decoding.

        Parameters:
            input: The input sequence(s). Can be either a single sequence, or a multidimensional array.

        Returns:
            output: The output sequence(s). Has the same shape as the input, with the last dimension multiplied by $m$.
        """

        # General minimum Euclidean distance hard demodulator.
        @vectorize
        def demodulate_hard(received: npt.NDArray[T]) -> npt.NDArray[np.integer]:
            return self.labeling[np.argmin(np.abs(self.constellation - received[0]))]

        input = np.asarray(input)
        received = input.reshape(*input.shape[:-1], -1, 1)
        hard_bits = demodulate_hard(received)
        output = hard_bits.reshape(*input.shape[:-1], -1)
        return output

    @abstractmethod
    def demodulate_soft(
        self, input: npt.ArrayLike, snr: float = 1.0
    ) -> npt.NDArray[np.floating]:
        r"""
        Demodulates one or more sequences of received points to their corresponding sequences of soft bits (L-values) using soft-decision decoding. The soft bits are the log-likelihood ratios of the bits, where positive values correspond to bit $\mathtt{0}$ and negative values correspond to bit $\mathtt{1}$.

        Parameters:
            input: The received sequence(s). Can be either a single sequence, or a multidimensional array.
            snr: The signal-to-noise ratio (SNR) of the channel. It should be a positive real number. The default value is `1.0`.

        Returns:
            output: The output sequence(s). Has the same shape as the input, with the last dimension multiplied by $m$.
        """

        @vectorize
        def demodulate_soft(received: npt.NDArray[T]) -> npt.NDArray[np.floating]:
            # Computes the L-values (LLR) of each bit. Assumes uniformly distributed bits.
            # See [SA15, eq. (3.50)].
            m = self.bits_per_symbol
            n0 = self.energy_per_symbol / snr
            received = received.ravel()

            # Precompute the distances and exponentials.
            distances = np.abs(received[:, np.newaxis] - self.constellation) ** 2
            exp_terms = np.exp(-distances / n0)

            soft_bits = np.empty(received.size * m, dtype=float)
            for bit_index in range(m):
                i0 = [i for i, q in enumerate(self.labeling) if q[bit_index] == 0]
                i1 = [i for i, q in enumerate(self.labeling) if q[bit_index] == 1]
                p0 = np.sum(exp_terms[:, i0], axis=1)
                p1 = np.sum(exp_terms[:, i1], axis=1)
                with np.errstate(divide="ignore"):
                    soft_bits[bit_index::m] = np.log(p0) - np.log(p1)
            return soft_bits

        input = np.asarray(input)
        received = input.reshape(*input.shape[:-1], -1, 1)
        soft_bits = demodulate_soft(received)
        output = soft_bits.reshape(*input.shape[:-1], -1)
        return output



from dataclasses import dataclass, field
from functools import cached_property

import numpy as np
import numpy.typing as npt

from .._util.information_theory import (
    PMF,
    LogBase,
    assert_is_probability,
    binary_entropy,
)
from . import base


@dataclass
class BinaryErasureChannel(base.DiscreteMemorylessChannel):
    r"""
    Binary erasure channel (BEC). It is a [discrete memoryless channel](/ref/DiscreteMemorylessChannel) with input alphabet $\mathcal{X} = \\{ 0, 1 \\}$ and output alphabet $\mathcal{Y} = \\{ 0, 1, 2 \\}$. The channel is characterized by a parameter $\epsilon$, called the *erasure probability*. With probability $1 - \epsilon$, the output symbol is identical to the input symbol, and with probability $\epsilon$, the output symbol is replaced by an erasure symbol (denoted by $2$). For more details, see <cite>CT06, Sec. 7.1.5</cite>.

    Attributes:
        erasure_probability: The channel erasure probability $\epsilon$. Must satisfy $0 \leq \epsilon \leq 1$. Default value is `0.0`, which corresponds to a noiseless channel.
    """

    erasure_probability: float = 0.0
    rng: np.random.Generator = field(default=np.random.default_rng(), repr=False)

    def __post_init__(self) -> None:
        assert_is_probability(self.erasure_probability)

    @cached_property
    def input_cardinality(self) -> int:
        r"""
        For the BEC, it is given by $|\mathcal{X}| = 2$.
        """
        return 2

    @cached_property
    def output_cardinality(self) -> int:
        r"""
        For the BEC, it is given by $|\mathcal{Y}| = 3$.
        """
        return 3

    @cached_property
    def transition_matrix(self) -> npt.NDArray[np.floating]:
        r"""
        For the BEC, it is given by
        $$
            p_{Y \mid X} =
            \begin{bmatrix}
                1 - \epsilon & 0 & \epsilon \\\\
                0 & 1 - \epsilon & \epsilon
            \end{bmatrix}.
        $$

        Examples:
            >>> bec = komm.BinaryErasureChannel(0.2)
            >>> bec.transition_matrix
            array([[0.8, 0. , 0.2],
                   [0. , 0.8, 0.2]])
        """
        epsilon = self.erasure_probability
        return np.array([[1 - epsilon, 0, epsilon], [0, 1 - epsilon, epsilon]])

    def mutual_information(
        self, input_pmf: npt.ArrayLike, base: LogBase = 2.0
    ) -> float:
        r"""
        For the BEC, it is given by
        $$
            \mathrm{I}(X ; Y) = (1 - \epsilon) \, \Hb(\pi),
        $$
        in bits, where $\pi = \Pr[X = 1]$, and $\Hb$ is the [binary entropy function](/ref/binary_entropy).

        Examples:
            >>> bec = komm.BinaryErasureChannel(0.2)
            >>> bec.mutual_information([0.45, 0.55]) # doctest: +FLOAT_CMP
            np.float64(0.7942195631902467)
        """
        input_pmf = PMF(input_pmf)
        epsilon = self.erasure_probability
        pi = input_pmf[1]
        return (1.0 - epsilon) * binary_entropy(pi) / np.log2(base)

    def capacity(self, base: LogBase = 2.0) -> float:
        r"""
        For the BEC, it is given by
        $$
            C = 1 - \epsilon,
        $$
        in bits.

        Examples:
            >>> bec = komm.BinaryErasureChannel(0.2)
            >>> bec.capacity()
            np.float64(0.8)
        """
        return (1.0 - self.erasure_probability) / np.log2(base)

    def __call__(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Examples:
            >>> rng = np.random.default_rng(seed=42)
            >>> bec = komm.BinaryErasureChannel(0.2, rng=rng)
            >>> bec([1, 1, 1, 0, 0, 0, 1, 0, 1, 0])
            array([1, 1, 1, 0, 2, 0, 1, 0, 2, 0])
        """
        epsilon = self.erasure_probability
        input = np.asarray(input)
        erasure_pattern = self.rng.random(input.shape) < epsilon
        output = np.copy(input)
        output[erasure_pattern] = 2
        return output



from functools import cache, cached_property

import numpy as np
import numpy.typing as npt

from .._algebra import BinaryPolynomial, BinaryPolynomialFraction
from .._finite_state_machine import FiniteStateMachine
from .._util.bit_operations import bits_to_int, int_to_bits


class ConvolutionalCode:
    r"""
    Binary convolutional code. It is characterized by a *matrix of feedforward polynomials* $\mathbf{P}(D)$, of shape $k \times n$, and (optionally) by a *vector of feedback polynomials* $\mathbf{q}(D)$, of length $k$. The element in row $i$ and column $j$ of $\mathbf{P}(D)$ is denoted by $p_{i,j}(D)$, and the element in position $i$ of $\mathbf{q}(D)$ is denoted by $q_i(D)$; they are [binary polynomials](/ref/BinaryPolynomial) in $D$. The parameters $k$ and $n$ are the number of input and output bits per block, respectively.

    The *transfer function matrix* (also known as *transform-domain generator matrix*) $\mathbf{G}(D)$ of the convolutional code, of shape $k \times n$, is such that the element in row $i$ and column $j$ is given by
    $$
        g_{i,j}(D) = \frac{p_{i,j}(D)}{q_{i}(D)},
    $$
    for $i \in [0 : k)$ and $j \in [0 : n)$.

    The *constraint lengths* of the code are defined by
    $$
        \nu_i = \max \\{ \deg p_{i,0}(D), \deg p_{i,1}(D), \ldots, \deg p_{i,n-1}(D), \deg q_i(D) \\},
    $$
    for $i \in [0 : k)$.

    The *overall constraint length* of the code is defined by
    $$
        \nu = \sum_{0 \leq i < k} \nu_i.
    $$

    The *memory order* of the code is defined by
    $$
        \mu = \max_{0 \leq i < k} \nu_i.
    $$

    For more details, see <cite>JZ15</cite> and <cite>LC04, Chs. 11, 12</cite>.

    Parameters:
        feedforward_polynomials: The matrix of feedforward polynomials $\mathbf{P}(D)$, which is a $k \times n$ matrix whose entries are either [binary polynomials](/ref/BinaryPolynomial) or integers to be converted to the former.

        feedback_polynomials: The vector of feedback polynomials $\mathbf{q}(D)$, which is a $k$-vector whose entries are either [binary polynomials](/ref/BinaryPolynomial) or integers to be converted to the former. The default value corresponds to no feedback, that is, $q_i(D) = 1$ for all $i \in [0 : k)$.

    Examples:
        The convolutional code with encoder depicted in the figure below has parameters $(n, k, \nu) = (2, 1, 6)$; its transfer function matrix is given by
        $$
            \mathbf{G}(D) =
            \begin{bmatrix}
                D^6 + D^3 + D^2 + D + 1  &  D^6 + D^5 + D^3 + D^2 + 1
            \end{bmatrix},
        $$
        yielding `feedforward_polynomials = [[0b1001111, 0b1101101]] = [[0o117, 0o155]] = [[79, 109]]`.

        <figure markdown>
          ![Convolutional encoder for (2, 1, 6) code.](/figures/cc_2_1_6.svg)
        </figure>

        >>> code = komm.ConvolutionalCode(feedforward_polynomials=[[0o117, 0o155]])
        >>> (code.num_output_bits, code.num_input_bits, code.overall_constraint_length)
        (2, 1, 6)

        The convolutional code with encoder depicted in the figure below has parameters $(n, k, \nu) = (3, 2, 7)$; its transfer function matrix is given by
        $$
            \mathbf{G}(D) =
            \begin{bmatrix}
                D^4 + D^3 + 1  &  D^4 + D^2 + D + 1  &  0 \\\\
                0  &  D^3 + D  &  D^3 + D^2 + 1
            \end{bmatrix},
        $$
        yielding `feedforward_polynomials = [[0b11001, 0b10111, 0b00000], [0b0000, 0b1010, 0b1101]] = [[0o31, 0o27, 0o00], [0o00, 0o12, 0o15]] = [[25, 23, 0], [0, 10, 13]]`.

        <figure markdown>
          ![Convolutional encoder for (3, 2, 7) code.](/figures/cc_3_2_7.svg)
        </figure>

        >>> code = komm.ConvolutionalCode(feedforward_polynomials=[[0o31, 0o27, 0o00], [0o00, 0o12, 0o15]])
        >>> (code.num_output_bits, code.num_input_bits, code.overall_constraint_length)
        (3, 2, 7)

        The convolutional code with feedback encoder depicted in the figure below has parameters $(n, k, \nu) = (2, 1, 4)$; its transfer function matrix is given by
        $$
            \mathbf{G}(D) =
            \begin{bmatrix}
                1  &  \dfrac{D^4 + D^3 + 1}{D^4 + D^2 + D + 1}
            \end{bmatrix},
        $$
        yielding `feedforward_polynomials = [[0b10111, 0b11001]] = [[0o27, 0o31]] = [[23, 25]]` and `feedback_polynomials = [0b10111] = [0o27] = [23]`.

        <figure markdown>
          ![Convolutional encoder for (2, 1, 4) feedback code.](/figures/cc_2_1_4_fb.svg)
        </figure>

        >>> code = komm.ConvolutionalCode(feedforward_polynomials=[[0o27, 0o31]], feedback_polynomials=[0o27])
        >>> (code.num_output_bits, code.num_input_bits, code.overall_constraint_length)
        (2, 1, 4)

    The tables below <cite>LC04, Sec. 12.3</cite> lists optimal convolutional codes with no feedback, for parameters $(n,k) = (2,1)$ and $(n,k) = (3,1)$, and small values of the overall constraint length $\nu$.

    | Parameters $(n, k, \nu)$ | Transfer function matrix $\mathbf{G}(D)$ |
    | :----------------------: | ---------------------------------------- |
    | $(2, 1, 1)$              | `[[0o1, 0o3]]`                           |
    | $(2, 1, 2)$              | `[[0o5, 0o7]]`                           |
    | $(2, 1, 3)$              | `[[0o13, 0o17]]`                         |
    | $(2, 1, 4)$              | `[[0o27, 0o31]]`                         |
    | $(2, 1, 5)$              | `[[0o53, 0o75]]`                         |
    | $(2, 1, 6)$              | `[[0o117, 0o155]]`                       |
    | $(2, 1, 7)$              | `[[0o247, 0o371]]`                       |
    | $(2, 1, 8)$              | `[[0o561, 0o753]]`                       |

    | Parameters $(n, k, \nu)$ | Transfer function matrix $\mathbf{G}(D)$ |
    | :----------------------: | ---------------------------------------- |
    | $(3, 1, 1)$              | `[[0o1, 0o3, 0o3]]`                      |
    | $(3, 1, 2)$              | `[[0o5, 0o7, 0o7]]`                      |
    | $(3, 1, 3)$              | `[[0o13, 0o15, 0o17]]`                   |
    | $(3, 1, 4)$              | `[[0o25, 0o33, 0o37]]`                   |
    | $(3, 1, 5)$              | `[[0o47, 0o53, 0o75]]`                   |
    | $(3, 1, 6)$              | `[[0o117, 0o127, 0o155]]`                |
    | $(3, 1, 7)$              | `[[0o255, 0o331, 0o367]]`                |
    | $(3, 1, 8)$              | `[[0o575, 0o623, 0o727]]`                |
    """

    def __init__(
        self,
        feedforward_polynomials: npt.ArrayLike,
        feedback_polynomials: npt.ArrayLike | None = None,
    ) -> None:
        vecBinaryPolynomial = np.vectorize(BinaryPolynomial)

        self.feedforward_polynomials = vecBinaryPolynomial(feedforward_polynomials)

        if feedback_polynomials is None:
            self.feedback_polynomials = vecBinaryPolynomial([0b1] * self.num_input_bits)
        else:
            self.feedback_polynomials = vecBinaryPolynomial(feedback_polynomials)

        self._feedback_polynomials_parameters = feedback_polynomials

    def __repr__(self) -> str:
        def vec_str(arr: npt.NDArray[np.object_]) -> str:
            return str(np.vectorize(str)(arr).tolist()).replace("'", "")

        args = f"feedforward_polynomials={vec_str(self.feedforward_polynomials)}"
        if self._feedback_polynomials_parameters is not None:
            args += f", feedback_polynomials={vec_str(self.feedback_polynomials)}"
        return f"{self.__class__.__name__}({args})"

    @cached_property
    def num_input_bits(self) -> int:
        r"""
        The number of input bits per block, $k$.
        """
        return self.feedforward_polynomials.shape[0]

    @cached_property
    def num_output_bits(self) -> int:
        r"""
        The number of output bits per block, $n$.
        """
        return self.feedforward_polynomials.shape[1]

    @cached_property
    def transfer_function_matrix(self) -> npt.NDArray[np.object_]:
        r"""
        The transfer function matrix $\mathbf{G}(D)$ of the code. This is a $k \times n$ array of [binary polynomial fractions](/ref/BinaryPolynomialFraction).
        """
        transfer_function_matrix = np.empty_like(self.feedforward_polynomials)
        for i, j in np.ndindex(self.feedforward_polynomials.shape):
            p = BinaryPolynomialFraction(self.feedforward_polynomials[i, j])
            q = BinaryPolynomialFraction(self.feedback_polynomials[i])
            transfer_function_matrix[i, j] = p / q
        return transfer_function_matrix

    @cached_property
    def constraint_lengths(self) -> npt.NDArray[np.integer]:
        r"""
        The constraint lengths $\nu_i$ of the code, for $i \in [0 : k)$. This is a $k$-array of integers.
        """
        nus = np.empty(self.num_input_bits, dtype=int)
        for i in range(self.num_input_bits):
            ps = self.feedforward_polynomials[i]
            q = self.feedback_polynomials[i]
            nus[i] = max(np.amax([p.degree for p in ps]), q.degree)
        return nus

    @cached_property
    def overall_constraint_length(self) -> int:
        r"""
        The overall constraint length $\nu$ of the code.
        """
        return int(np.sum(self.constraint_lengths))

    @cached_property
    def memory_order(self) -> int:
        r"""
        The memory order $\mu$ of the code.
        """
        return int(np.max(self.constraint_lengths))

    @cache
    def finite_state_machine(self) -> FiniteStateMachine:
        r"""
        Returns the [finite-state machine](/ref/FiniteStateMachine) of the code, in direct form.

        Returns:
            The finite-state machine of the code.

        Examples:
            >>> code = komm.ConvolutionalCode(feedforward_polynomials=[[0b101, 0b111]])
            >>> code.finite_state_machine()
            FiniteStateMachine(next_states=[[0, 1], [2, 3], [0, 1], [2, 3]], outputs=[[0, 3], [2, 1], [3, 0], [1, 2]])
        """
        n = self.num_output_bits
        k = self.num_input_bits
        nu = self.overall_constraint_length

        x_indices = np.concatenate(([0], np.cumsum(self.constraint_lengths + 1)[:-1]))
        s_indices = np.setdiff1d(np.arange(k + nu, dtype=int), x_indices)

        ff_taps: list[npt.NDArray[np.integer]] = []
        for j in range(n):
            taps = np.concatenate([
                self.feedforward_polynomials[i, j].exponents() + x_indices[i]
                for i in range(k)
            ])
            ff_taps.append(taps)

        fb_taps: list[npt.NDArray[np.integer]] = []
        for i in range(k):
            taps = (
                BinaryPolynomial(0b1) + self.feedback_polynomials[i]
            ).exponents() + x_indices[i]
            fb_taps.append(taps)

        bits = np.empty(k + nu, dtype=int)
        next_states = np.empty((2**nu, 2**k), dtype=int)
        outputs = np.empty((2**nu, 2**k), dtype=int)

        for s, x in np.ndindex(2**nu, 2**k):
            bits[s_indices] = int_to_bits(s, width=nu)
            bits[x_indices] = int_to_bits(x, width=k)
            bits[x_indices] ^= [np.sum(bits[fb_taps[i]]) % 2 for i in range(k)]

            next_state_bits = bits[s_indices - 1]
            output_bits = [np.sum(bits[ff_taps[j]]) % 2 for j in range(n)]

            next_states[s, x] = bits_to_int(next_state_bits)
            outputs[s, x] = bits_to_int(output_bits)

        return FiniteStateMachine(next_states, outputs)

    @cache
    def state_space_representation(
        self,
    ) -> tuple[
        npt.NDArray[np.integer],
        npt.NDArray[np.integer],
        npt.NDArray[np.integer],
        npt.NDArray[np.integer],
    ]:
        r"""
        Returns the *state-space representation* of the code. Let
        $$
        \begin{aligned}
            \mathbf{u}_t & = (u_t^{(0)}, u_t^{(1)}, \ldots, u_t^{(k-1)}), \\\\
            \mathbf{v}_t & = (v_t^{(0)}, v_t^{(1)}, \ldots, v_t^{(n-1)}), \\\\
            \mathbf{s}_t & = (s_t^{(0)}, s_t^{(1)}, \ldots, s_t^{(\nu-1)}),
        \end{aligned}
        $$
        be the input block, output block, and state, respectively, all defined at time instant $t$. Then,
        $$
        \begin{aligned}
            \mathbf{s}\_{t+1} & = \mathbf{s}_t \mathbf{A} + \mathbf{u}_t \mathbf{B}, \\\\
            \mathbf{v}\_{t} & = \mathbf{s}_t \mathbf{C} + \mathbf{u}\_t \mathbf{D},
        \end{aligned}
        $$
        where $\mathbf{A}$ is the $\nu \times \nu$ *state matrix*, $\mathbf{B}$ is the $k \times \nu$ *control matrix*, $\mathbf{C}$ is the $\nu \times n$ *observation matrix*, and $\mathbf{D}$ is the $k \times n$ *transition matrix*. They are all binary matrices. For more details, see <cite>WBR01</cite>.

        Returns:
            The state matrix $\mathbf{A}$, control matrix $\mathbf{B}$, observation matrix $\mathbf{C}$, and transition matrix $\mathbf{D}$ of the code.

        Examples:
            >>> code = komm.ConvolutionalCode(feedforward_polynomials=[[0b101, 0b111]])
            >>> state_matrix, control_matrix, observation_matrix, transition_matrix = code.state_space_representation()
            >>> state_matrix
            array([[0, 1],
                   [0, 0]])
            >>> control_matrix
            array([[1, 0]])
            >>> observation_matrix
            array([[0, 1],
                   [1, 1]])
            >>> transition_matrix
            array([[1, 1]])
        """
        k = self.num_input_bits
        n = self.num_output_bits
        nu = self.overall_constraint_length
        fsm = self.finite_state_machine()

        state_matrix = np.empty((nu, nu), dtype=int)
        observation_matrix = np.empty((nu, n), dtype=int)
        for i in range(nu):
            s0 = 2**i
            state_matrix[i, :] = int_to_bits(fsm.next_states[s0, 0], width=nu)
            observation_matrix[i, :] = int_to_bits(fsm.outputs[s0, 0], width=n)

        control_matrix = np.empty((k, nu), dtype=int)
        transition_matrix = np.empty((k, n), dtype=int)
        for i in range(k):
            x = 2**i
            control_matrix[i, :] = int_to_bits(fsm.next_states[0, x], width=nu)
            transition_matrix[i, :] = int_to_bits(fsm.outputs[0, x], width=n)

        return state_matrix, control_matrix, observation_matrix, transition_matrix



from dataclasses import dataclass, field
from functools import cached_property
from typing import Literal

import numpy as np
import numpy.typing as npt

from . import base


@dataclass(eq=False)
class UniformQuantizer(base.ScalarQuantizer):
    r"""
    Uniform scalar quantizer. It is a [scalar quantizer](/ref/ScalarQuantizer) in which the separation between levels is constant, $\Delta$, and the thresholds are the mid-point between adjacent levels. For more details, see <cite>Say06, Sec. 9.4</cite>.

    Attributes:
        num_levels: The number of quantization levels $L$. It must be greater than $1$.

        input_range: The range $(x_\mathrm{min}, x_\mathrm{max})$ of the input signal. The default is $(-1.0, 1.0)$.

        choice: The choice for the uniform quantizer. Must be either `'mid-riser'` or `'mid-tread'`. The default is `'mid-riser'`.

    Examples:
        >>> quantizer = komm.UniformQuantizer(
        ...     num_levels=4,
        ...     input_range=(-1.0, 1.0),
        ...     choice='mid-riser',
        ... )
        >>> quantizer.levels
        array([-0.75, -0.25,  0.25,  0.75])
        >>> quantizer.thresholds
        array([-0.5,  0. ,  0.5])

        >>> quantizer = komm.UniformQuantizer(
        ...     num_levels=4,
        ...     input_range=(-1.0, 1.0),
        ...     choice='mid-tread',
        ... )
        >>> quantizer.levels
        array([-1. , -0.5,  0. ,  0.5])
        >>> quantizer.thresholds
        array([-0.75, -0.25,  0.25])

        >>> quantizer = komm.UniformQuantizer(
        ...     num_levels=4,
        ...     input_range=(0.0, 1.0),
        ...     choice='mid-tread',
        ... )
        >>> quantizer.levels
        array([0.  , 0.25, 0.5 , 0.75])
        >>> quantizer.thresholds
        array([0.125, 0.375, 0.625])
    """

    num_levels: int
    input_range: tuple[float, float] = field(default=(-1.0, 1.0))
    choice: Literal["mid-riser", "mid-tread"] = field(default="mid-riser")

    def __post_init__(self) -> None:
        if not self.num_levels > 1:
            raise ValueError("'num_levels' must be greater than 1")
        if not self.choice in ["mid-riser", "mid-tread"]:
            raise ValueError("'choice' must be either 'mid-riser' or 'mid-tread'")

    @cached_property
    def quantization_step(self) -> float:
        r"""
        The quantization step $\Delta$.

        Examples:
            >>> quantizer = komm.UniformQuantizer(num_levels=4, input_range=(-1.0, 1.0))
            >>> quantizer.quantization_step
            0.5
        """
        x_min, x_max = self.input_range
        delta = (x_max - x_min) / self.num_levels
        return delta

    @cached_property
    def levels(self) -> npt.NDArray[np.floating]:
        r"""
        Examples:
            >>> quantizer = komm.UniformQuantizer(num_levels=4, input_range=(-1.0, 1.0))
            >>> quantizer.levels
            array([-0.75, -0.25,  0.25,  0.75])
        """
        num = self.num_levels
        x_min, x_max = self.input_range
        delta = (x_max - x_min) / num
        endpoint = (num % 2 == 0) if self.choice == "mid-riser" else (num % 2 == 1)
        min_level = x_min + (delta / 2) * endpoint
        max_level = x_max - (delta / 2) * endpoint
        return np.linspace(min_level, max_level, num=num, endpoint=endpoint)

    @cached_property
    def thresholds(self) -> npt.NDArray[np.floating]:
        r"""
        Examples:
            >>> quantizer = komm.UniformQuantizer(num_levels=4, input_range=(-1.0, 1.0))
            >>> quantizer.thresholds
            array([-0.5,  0. ,  0.5])
        """
        return (self.levels + self.quantization_step / 2)[:-1]

    def digitize(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Examples:
            >>> quantizer = komm.UniformQuantizer(num_levels=4, input_range=(-1.0, 1.0))
            >>> quantizer.digitize([-0.6, 0.2, 0.8])
            array([0, 2, 3])
        """
        input = np.asarray(input, dtype=float)
        output = np.digitize(input, self.thresholds, right=True)
        return output

    def quantize(self, input: npt.ArrayLike) -> npt.NDArray[np.floating]:
        r"""
        Examples:
            >>> quantizer = komm.UniformQuantizer(num_levels=4, input_range=(-1.0, 1.0))
            >>> quantizer.quantize([-0.6, 0.2, 0.8])
            array([-0.75,  0.25,  0.75])
        """
        input = np.array(input, dtype=float, ndmin=1)
        delta = self.quantization_step
        if self.choice == "mid-riser":
            quantized = delta * (np.floor(input / delta) + 0.5)
        else:
            quantized = delta * np.floor(input / delta + 0.5)
        output = np.clip(quantized, a_min=self.levels[0], a_max=self.levels[-1])
        return output



from dataclasses import dataclass

import numpy as np
import numpy.typing as npt

from .._util.bit_operations import bits_to_int, int_to_bits
from .ConvolutionalCode import ConvolutionalCode


@dataclass
class ConvolutionalStreamEncoder:
    r"""
    Convolutional stream encoder. Encode a bit stream using a given [convolutional code](/ref/ConvolutionalCode). The internal state of the encoder is maintained across each call.

    Attributes:
        convolutional_code: The convolutional code.
        state: The current state of the encoder. The default value is `0`.
    """

    convolutional_code: ConvolutionalCode
    state: int = 0

    def __call__(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Parameters:
            input: The bit sequence to be encoded.

        Returns:
            output: The encoded bit sequence.

        Examples:
            >>> convolutional_code = komm.ConvolutionalCode([[0o7, 0o5]])
            >>> encoder = komm.ConvolutionalStreamEncoder(convolutional_code)
            >>> encoder([1, 1, 1, 1])
            array([1, 1, 0, 1, 1, 0, 1, 0])
            >>> encoder([1, 1, 1, 1])
            array([1, 0, 1, 0, 1, 0, 1, 0])
        """
        input = np.asarray(input)
        n = self.convolutional_code.num_output_bits
        k = self.convolutional_code.num_input_bits
        fsm = self.convolutional_code.finite_state_machine()
        output_sequence, self.state = fsm.process(
            input_sequence=bits_to_int(input.reshape(-1, k)),
            initial_state=self.state,
        )
        output = int_to_bits(output_sequence, width=n).ravel()
        return output



import numpy as np
import numpy.typing as npt
from tqdm import tqdm

ArrayInt = npt.NDArray[np.integer]


def _pivot(row: ArrayInt) -> int:
    f_list = np.flatnonzero(row)
    return f_list[0] if f_list.size > 0 else row.size


def rref(matrix: npt.ArrayLike) -> ArrayInt:
    r"""
    Computes the row-reduced echelon form of a matrix in $\ZZ_2$.

    Parameters:
        matrix: The matrix to reduce. Its elements must be `0` or `1`.

    Returns:
        reduced: The row-reduced echelon form of the matrix.

    Examples:
        >>> rref([[1, 1, 1], [1, 1, 0], [0, 0, 1]])
        array([[1, 1, 0],
               [0, 0, 1],
               [0, 0, 0]])

    References:
        [1] https://gist.github.com/rgov/1499136
    """
    reduced = np.asarray(matrix, dtype=int)
    n_rows, n_cols = reduced.shape
    for r in tqdm(range(n_rows), desc="Row-reducing matrix", unit="row", delay=2.5):
        # Choose the pivot
        possible_pivots = [_pivot(row) for row in reduced[r:]]
        p = np.argmin(possible_pivots) + r
        # Swap rows
        reduced[r], reduced[p] = reduced[p].copy(), reduced[r].copy()
        # Pivot column
        f = _pivot(reduced[r])
        if f >= n_cols:
            continue
        # Subtract the row from others
        for i in range(n_rows):
            if i != r and reduced[i, f] != 0:
                reduced[i] = (reduced[i] + reduced[r]) % 2
    return reduced


def xrref(matrix: npt.ArrayLike) -> tuple[ArrayInt, ArrayInt, ArrayInt]:
    r"""
    Computes the row-reduced echelon form of a matrix in $\ZZ_2$. Returns the row transformation matrix, the reduced matrix, and the pivot indices. The relation between the input matrix `matrix` and the reduced matrix `reduced` is given by the equation `reduced = row_transform @ matrix`.

    Parameters:
        matrix: The matrix to reduce. Its elements must be `0` or `1`.

    Returns:
        row_transform: The row transformation matrix.
        reduced: The row-reduced echelon form of the matrix.
        pivots: The pivot indices.

    Examples:
        >>> matrix = np.array([[1, 1, 1], [1, 1, 0], [0, 0, 1]])
        >>> row_transform, reduced, pivots = xrref(matrix)
        >>> row_transform
        array([[0, 1, 0],
               [0, 0, 1],
               [1, 1, 1]])
        >>> reduced
        array([[1, 1, 0],
               [0, 0, 1],
               [0, 0, 0]])
        >>> pivots
        array([0, 2])
        >>> np.array_equal((row_transform @ matrix) % 2, reduced)
        True
    """
    matrix = np.asarray(matrix, dtype=int)
    n_rows, n_cols = matrix.shape
    eye = np.eye(n_rows, dtype=int)
    augmented = np.concatenate((matrix, eye), axis=1)
    augmented_rref = rref(augmented)

    reduced = augmented_rref[:, :n_cols]
    row_transform = augmented_rref[:, n_cols:]

    pivots: list[int] = []
    for j, col in enumerate(reduced.T):
        if len(pivots) >= n_rows:
            break
        if np.array_equal(col, eye[len(pivots)]):
            pivots.append(j)

    return row_transform, reduced, np.array(pivots)


def rank(matrix: npt.ArrayLike) -> int:
    r"""
    Computes the rank of a matrix in $\ZZ_2$.

    Parameters:
        matrix: The matrix. Its elements must be `0` or `1`.

    Returns:
        rank: The rank of the matrix.

    Examples:
        >>> rank([[1, 1, 1], [1, 1, 0], [0, 0, 1]])
        2
    """
    _, _, pivots = xrref(matrix)
    return len(pivots)


def pseudo_inverse(matrix: npt.ArrayLike) -> ArrayInt:
    r"""
    Computes a pseudo inverse of a matrix in $\ZZ_2$.

    Parameters:
        matrix: The matrix. Its elements must be `0` or `1`.

    Returns:
        p_inverse: A pseudo inverse of the matrix.

    Examples:
        >>> matrix = np.array([[1, 0, 1], [1, 1, 1]])
        >>> p_inverse = pseudo_inverse(matrix)
        >>> p_inverse
        array([[1, 0],
               [1, 1],
               [0, 0]])
        >>> (matrix @ p_inverse) % 2
        array([[1, 0],
               [0, 1]])

        >>> matrix = np.array([[1, 1, 1], [1, 1, 0], [0, 0, 1]])
        >>> p_inverse = pseudo_inverse(matrix)
        >>> p_inverse
        array([[0, 1, 0],
               [0, 0, 0],
               [0, 0, 1]])
        >>> np.array_equal((matrix @ p_inverse @ matrix) % 2, matrix)
        True
        >>> np.array_equal((p_inverse @ matrix @ p_inverse) % 2, p_inverse)
        True
    """
    row_transform, reduced, pivots = xrref(matrix)
    reduced_inverse = np.zeros_like(reduced.T)
    if pivots.size == 0:
        return reduced_inverse
    reduced_inverse[pivots] = np.eye(pivots.size, reduced.shape[0])
    p_inverse = np.dot(reduced_inverse, row_transform) % 2
    return p_inverse


def null_matrix(matrix: npt.ArrayLike) -> ArrayInt:
    r"""
    Computes a null matrix of a matrix in $\ZZ_2$.
    """
    _, reduced, pivots = xrref(matrix)
    n_rows, n_cols = reduced.shape
    null = np.empty((n_cols - n_rows, n_cols), dtype=int)
    not_pivots = np.setdiff1d(range(n_cols), pivots)
    null[:, not_pivots] = np.eye(n_cols - n_rows, dtype=int)
    null[:, pivots] = reduced[:, not_pivots].T
    return null



from functools import cached_property
from typing import Literal

import numpy as np
import numpy.typing as npt

from . import base
from .constellations import constellation_apsk
from .labelings import get_labeling


class APSKModulation(base.Modulation[np.complexfloating]):
    r"""
    Amplitude- and phase-shift keying (APSK) modulation. It is a complex [modulation scheme](/ref/Modulation) in which the constellation is the union (concatenation) of component [PSK](/ref/PSKModulation) constellations, called *rings*. More precisely, consider $K$ rings $\mathbf{X}_k$, for $k \in [0 : K)$, where the $k$-th ring has order $M_k$, amplitude $A_k$, and phase offset $\phi_k$. The $i$-th constellation symbol of the $k$-th ring is given by
    $$
        x\_{k,i} = A_k \exp \left( \mathrm{j} \frac{2 \pi i}{M_k} \right) \exp(\mathrm{j} \phi_k),
        \quad k \in [0 : K),
        \quad i \in [0 : M_k).
    $$
    The resulting APSK constellation is therefore given by
    $$
        \mathbf{X} = \begin{bmatrix}
            \mathbf{X}_0 \\\\
            \vdots \\\\
            \mathbf{X}\_{K-1}
        \end{bmatrix},
    $$
    which has order $M = M_0 + M_1 + \cdots + M\_{K-1}$. The order $M_k$ of each ring need not be a power of $2$; however, the order $M$ of the constructed APSK modulation must be.

    Parameters:
        orders: A $K$-tuple with the orders $M_k$ of each ring, for $k \in [0 : K)$. The sum $M_0 + M_1 + \cdots + M_{K-1}$ must be a power of $2$.

        amplitudes: A $K$-tuple with the amplitudes $A_k$ of each ring, for $k \in [0 : K)$.

        phase_offsets: A $K$-tuple with the phase offsets $\phi_k$ of each ring, for $k \in [0 : K)$. If specified as a single float $\phi$, then it is assumed that $\phi_k = \phi$ for all $k \in [0 : K)$. The default value is `0.0`.

        labeling: The binary labeling of the modulation. Can be specified either as a 2D-array of integers (see [base class](/ref/Modulation) for details), or as a string. In the latter case, the string must be equal to `'natural'`. The default value is `'natural'`.

    Examples:

    1. The $8$-APSK modulation with $(M_0, M_1) = (4, 4)$, $(A_0, A_1) = (1, 2)$, and $(\phi_0, \phi_1) = (0, 0)$ is depicted below.
        <figure markdown>
        ![(4,4)-APSK modulation.](/figures/apsk_4_4.svg)
        </figure>

            >>> apsk = komm.APSKModulation(orders=(4, 4), amplitudes=(1.0, 2.0))
            >>> apsk.constellation.round(3)
            array([ 1.+0.j,  0.+1.j, -1.+0.j, -0.-1.j,
                    2.+0.j,  0.+2.j, -2.+0.j, -0.-2.j])

    1. The $16$-APSK modulation with $(M_0, M_1) = (8, 8)$, $(A_0, A_1) = (1, 2)$, and $(\phi_0, \phi_1) = (0, \pi/8)$ is depicted below.
        <figure markdown>
        ![(8,8)-APSK modulation.](/figures/apsk_8_8.svg)
        </figure>

            >>> apsk = komm.APSKModulation(
            ...     orders=(8, 8),
            ...     amplitudes=(1.0, 2.0),
            ...     phase_offsets=(0.0, np.pi/8)
            ... )
            >>> apsk.constellation.round(3)
            array([ 1.   +0.j   ,  0.707+0.707j,  0.   +1.j   , -0.707+0.707j,
                   -1.   +0.j   , -0.707-0.707j, -0.   -1.j   ,  0.707-0.707j,
                    1.848+0.765j,  0.765+1.848j, -0.765+1.848j, -1.848+0.765j,
                   -1.848-0.765j, -0.765-1.848j,  0.765-1.848j,  1.848-0.765j])

    1. The $16$-APSK modulation with $(M_0, M_1) = (4, 12)$, $(A_0, A_1) = (\sqrt{2}, 3)$, and $(\phi_0, \phi_1) = (\pi/4, 0)$ is depicted below.
        <figure markdown>
        ![(4,12)-APSK modulation.](/figures/apsk_4_12.svg)
        </figure>

            >>> apsk = komm.APSKModulation(
            ...     orders=(4, 12),
            ...     amplitudes=(np.sqrt(2), 3.0),
            ...     phase_offsets=(np.pi/4, 0.0)
            ... )
            >>> apsk.constellation.round(3)
            array([ 1.   +1.j   , -1.   +1.j   , -1.   -1.j   ,  1.   -1.j   ,
                    3.   +0.j   ,  2.598+1.5j  ,  1.5  +2.598j,  0.   +3.j   ,
                   -1.5  +2.598j, -2.598+1.5j  , -3.   +0.j   , -2.598-1.5j  ,
                   -1.5  -2.598j, -0.   -3.j   ,  1.5  -2.598j,  2.598-1.5j  ])
    """

    def __init__(
        self,
        orders: tuple[int, ...],
        amplitudes: tuple[float, ...],
        phase_offsets: float | tuple[float, ...] = 0.0,
        labeling: Literal["natural"] | npt.ArrayLike = "natural",
    ) -> None:
        self._orders = orders
        self._amplitudes = amplitudes
        if isinstance(phase_offsets, tuple):
            self._phase_offsets = phase_offsets
        else:
            self._phase_offsets = (phase_offsets,) * len(orders)
        self._labeling = labeling
        super()._validate_parameters()

    def __repr__(self) -> str:
        if isinstance(self._labeling, str):
            labeling_repr = repr(self._labeling)
        else:
            labeling_repr = self.labeling.tolist()
        args = ", ".join([
            f"orders={self._orders}",
            f"amplitudes={self._amplitudes}",
            f"phase_offsets={self._phase_offsets}",
            f"labeling={labeling_repr}",
        ])
        return f"{self.__class__.__name__}({args})"

    @cached_property
    def constellation(self) -> npt.NDArray[np.complexfloating]:
        r"""
        Examples:
            >>> apsk = komm.APSKModulation(orders=(4, 4), amplitudes=(1.0, 2.0))
            >>> apsk.constellation.round(3)
            array([ 1.+0.j,  0.+1.j, -1.+0.j, -0.-1.j,
                    2.+0.j,  0.+2.j, -2.+0.j, -0.-2.j])
        """
        return constellation_apsk(self._orders, self._amplitudes, self._phase_offsets)

    @cached_property
    def labeling(self) -> npt.NDArray[np.integer]:
        r"""
        Examples:
            >>> apsk = komm.APSKModulation(orders=(4, 4), amplitudes=(1.0, 2.0))
            >>> apsk.labeling
            array([[0, 0, 0],
                   [0, 0, 1],
                   [0, 1, 0],
                   [0, 1, 1],
                   [1, 0, 0],
                   [1, 0, 1],
                   [1, 1, 0],
                   [1, 1, 1]])
        """
        return get_labeling(self._labeling, ("natural",), sum(self._orders))

    @cached_property
    def inverse_labeling(self) -> dict[tuple[int, ...], int]:
        r"""
        Examples:
            >>> apsk = komm.APSKModulation(orders=(4, 4), amplitudes=(1.0, 2.0))
            >>> apsk.inverse_labeling
            {(0, 0, 0): 0,
             (0, 0, 1): 1,
             (0, 1, 0): 2,
             (0, 1, 1): 3,
             (1, 0, 0): 4,
             (1, 0, 1): 5,
             (1, 1, 0): 6,
             (1, 1, 1): 7}
        """
        return super().inverse_labeling

    @cached_property
    def order(self) -> int:
        r"""
        Examples:
            >>> apsk = komm.APSKModulation(orders=(4, 4), amplitudes=(1.0, 2.0))
            >>> apsk.order
            8
        """
        return sum(self._orders)

    @cached_property
    def bits_per_symbol(self) -> int:
        r"""
        Examples:
            >>> apsk = komm.APSKModulation(orders=(4, 4), amplitudes=(1.0, 2.0))
            >>> apsk.bits_per_symbol
            3
        """
        return super().bits_per_symbol

    @cached_property
    def energy_per_symbol(self) -> float:
        r"""
        Examples:
            >>> apsk = komm.APSKModulation(orders=(4, 4), amplitudes=(1.0, 2.0))
            >>> apsk.energy_per_symbol
            2.5
        """
        return super().energy_per_symbol

    @cached_property
    def energy_per_bit(self) -> float:
        r"""
        Examples:
            >>> apsk = komm.APSKModulation(orders=(4, 4), amplitudes=(1.0, 2.0))
            >>> apsk.energy_per_bit
            0.8333333333333334
        """
        return super().energy_per_bit

    @cached_property
    def symbol_mean(self) -> complex:
        r"""
        Examples:
            >>> apsk = komm.APSKModulation(orders=(4, 4), amplitudes=(1.0, 2.0))
            >>> apsk.symbol_mean
            0j
        """
        return 0j

    @cached_property
    def minimum_distance(self) -> float:
        r"""
        Examples:
            >>> apsk = komm.APSKModulation(orders=(4, 4), amplitudes=(1.0, 2.0))
            >>> apsk.minimum_distance
            1.0
        """
        return super().minimum_distance

    def modulate(self, input: npt.ArrayLike) -> npt.NDArray[np.complexfloating]:
        r"""
        Examples:
            >>> apsk = komm.APSKModulation(orders=(4, 4), amplitudes=(1.0, 2.0))
            >>> apsk.modulate([0, 0, 0, 0, 1, 1, 0, 0, 0]).round(3)
            array([ 1.+0.j, -0.-1.j,  1.+0.j])
        """
        return super().modulate(input)

    def demodulate_hard(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        return super().demodulate_hard(input)

    def demodulate_soft(
        self, input: npt.ArrayLike, snr: float = 1.0
    ) -> npt.NDArray[np.floating]:
        return super().demodulate_soft(input, snr)



from dataclasses import dataclass
from functools import cache
from math import comb

import numpy as np
import numpy.typing as npt

from .._util.docs import mkdocstrings
from .BlockCode import BlockCode


@mkdocstrings(filters=["!.*"])
@dataclass(eq=False)
class RepetitionCode(BlockCode):
    r"""
    Repetition code. For a given length $n \geq 1$, it is the [linear block code](/ref/BlockCode) whose only two codewords are $00 \cdots 0$ and $11 \cdots 1$. The repetition code has the following parameters:

    - Length: $n$
    - Dimension: $k = 1$
    - Redundancy: $m = n - 1$
    - Minimum distance: $d = n$

    Notes:
        - Its dual is the [single parity-check code](/ref/SingleParityCheckCode).

    Attributes:
        n: The length $n$ of the code. Must be a positive integer.

    Examples:
        >>> code = komm.RepetitionCode(5)
        >>> (code.length, code.dimension, code.redundancy)
        (5, 1, 4)
        >>> code.generator_matrix
        array([[1, 1, 1, 1, 1]])
        >>> code.check_matrix
        array([[1, 1, 0, 0, 0],
               [1, 0, 1, 0, 0],
               [1, 0, 0, 1, 0],
               [1, 0, 0, 0, 1]])
        >>> code.minimum_distance()
        5

        >>> code = komm.RepetitionCode(16)
        >>> code.codeword_weight_distribution()
        array([1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1])
        >>> code.coset_leader_weight_distribution()
        array([    1,    16,   120,   560,  1820,  4368,  8008, 11440,  6435,
                   0,     0,     0,     0,     0,     0,     0,     0])
    """

    n: int

    def __post_init__(self):
        if not self.n >= 1:
            raise ValueError("n must be a positive integer")
        super().__init__(generator_matrix=np.ones((1, self.n), dtype=int))

    @cache
    def minimum_distance(self) -> int:
        return self.n

    @cache
    def coset_leader_weight_distribution(self) -> npt.NDArray[np.integer]:
        n = self.n
        coset_leader_weight_distribution = np.zeros(n + 1, dtype=int)
        for w in range((n + 1) // 2):
            coset_leader_weight_distribution[w] = comb(n, w)
        if n % 2 == 0:
            coset_leader_weight_distribution[n // 2] = comb(n, n // 2) // 2
        return coset_leader_weight_distribution



from dataclasses import dataclass
from functools import cached_property

import numpy as np
import numpy.typing as npt

from . import base


@dataclass
class RectangularPulse(base.Pulse):
    r"""
    Rectangular pulse. It is a [pulse](/ref/Pulse) with waveform given by
    $$
        h(t) =
        \begin{cases}
            1, & 0 \leq t < w, \\\\
            0, & \text{otherwise}.
        \end{cases}
        = \mathrm{rect}\left(\frac{t}{w}\right),
    $$
    where $w$ is the *width* of the pulse, which must satisfy $0 \leq w \leq 1$. Its spectrum is given by
    $$
        \hat{h}(f) = w \, \operatorname{sinc}(w f).
    $$

    The rectangular pulse is depicted below for $w = 1$, and for $w = 0.5$.

    <div class="centered" markdown>
      <span>
        ![Rectangular NRZ pulse.](/figures/pulse_rectangular_nrz.svg)
      </span>
      <span>
        ![Rectangular RZ pulse.](/figures/pulse_rectangular_rz.svg)
      </span>
    </div>

    Notes:
        - For $w = 1$ it is also called the _NRZ pulse_.
        - For $w = 0.5$ it is also called the _halfway RZ pulse_.

    Attributes:
        width: The width $w$ of the pulse. Must satisfy $0 \leq w \leq 1$. The default value is `1.0`.

    Examples:
        >>> pulse = komm.RectangularPulse(width=1.0)  # NRZ pulse
        >>> pulse.waveform([-0.50, -0.25,  0.00,  0.25,  0.50,  0.75,  1.00])
        array([0., 0., 1., 1., 1., 1., 0.])
        >>> pulse.spectrum([-0.75, -0.50, -0.25,  0.00,  0.25,  0.50,  0.75]).round(4)
        array([0.3001, 0.6366, 0.9003, 1.    , 0.9003, 0.6366, 0.3001])

        >>> pulse = komm.RectangularPulse(width=0.5)  # Halfway RZ pulse
        >>> pulse.waveform([-0.50, -0.25,  0.00,  0.25,  0.50,  0.75,  1.00])
        array([0., 0., 1., 1., 0., 0., 0.])
        >>> pulse.spectrum([-0.75, -0.50, -0.25,  0.00,  0.25,  0.50,  0.75]).round(4)
        array([0.3921, 0.4502, 0.4872, 0.5   , 0.4872, 0.4502, 0.3921])
    """

    width: float = 1.0

    def waveform(self, t: npt.ArrayLike) -> npt.NDArray[np.floating]:
        t = np.asarray(t)
        return 1.0 * (0.0 <= t) * (t < self.width)

    def spectrum(self, f: npt.ArrayLike) -> npt.NDArray[np.floating]:
        f = np.asarray(f)
        return self.width * np.sinc(self.width * f)

    @cached_property
    def support(self) -> tuple[float, float]:
        return (0.0, self.width)



import numpy as np
import numpy.typing as npt


def extended_parity_submatrix(
    parity_submatrix: npt.NDArray[np.integer],
) -> npt.NDArray[np.integer]:
    last_column = (1 + np.sum(parity_submatrix, axis=1)) % 2
    extended_parity_submatrix = np.hstack([parity_submatrix, last_column[np.newaxis].T])
    return extended_parity_submatrix
